# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

import numpy as np
from os.path import split
from abc import ABC, abstractmethod
from uncertainties import unumpy
from scipy.constants import physical_constants
from scipy.integrate import simps
from ase.geometry import get_distances

from pyiron_atomistics import Project
from pyiron_atomistics.atomistics.job.interactive import GenericInteractive
from pyiron_atomistics.lammps.lammps import LammpsInteractive
from pyiron_contrib.protocol.generic import PrimitiveVertex
from pyiron_contrib.protocol.utils import Pointer
from pyiron_contrib.protocol.utils import ensure_iterable
from pyiron_contrib.protocol.math import welford_online

KB = physical_constants['Boltzmann constant in eV/K'][0]
EV_TO_U_ANGSQ_PER_FSSQ = 0.00964853322
# https://www.wolframalpha.com/input/?i=1+eV+in+u+*+%28angstrom%2Ffs%29%5E2
U_ANGSQ_PER_FSSQ_TO_EV = 1. / EV_TO_U_ANGSQ_PER_FSSQ
EV_PER_ANGCUB_TO_GPA = 160.21766208  # eV/A^3 to GPa
GPA_TO_BAR = 1e4

"""
Primitive vertices which have only one outbound execution edge.
"""

__author__ = "Liam Huber, Raynol Dsouza, Dominik Noeger"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "20 July, 2019"


class BuildMixingPairs(PrimitiveVertex):
    """
    Builds an array of mixing parameters [lambda, (1-lambda)], and also finds the deltas between consecutive
        lambdas.
    Input attributes:
        n_lambdas (int): How many mixing pairs to create. (Default is 5.)
        custom_lambdas (list/numpy.ndarray): The individual lambda values to use for the first member of each pair.
            (Default is None.)
    Output attributes:
        lambda_pairs (numpy.ndarray): The (`n_lambdas`, 2)-shaped array of mixing pairs.
        delta_lambdas (numpy.ndarray): The delta between two consecutive lambdas. The end deltas are
            halved.
    """

    def __init__(self, name=None):
        super(BuildMixingPairs, self).__init__(name=name)
        self.input.default.n_lambdas = 5
        self.input.default.custom_lambdas = None

    def command(self, n_lambdas, custom_lambdas):

        if custom_lambdas is not None:
            lambdas = np.array(custom_lambdas)
        else:
            lambdas = np.linspace(0, 1, num=n_lambdas)

        delta_lambdas = np.gradient(lambdas)
        delta_lambdas[0] = delta_lambdas[0] / 2
        delta_lambdas[-1] = delta_lambdas[-1] / 2

        return {
                'lambda_pairs': np.array([lambdas, 1 - lambdas]).T,
                'delta_lambdas': delta_lambdas
            }


class Counter(PrimitiveVertex):
    """
    Increments by one at each execution. Can be made to increment from a specific value.
    Input attributes:
        add_counts (int): A specific value from which to increment. (Default is 0.)
    Output attributes:
        n_counts (int): How many executions have passed. (Default is 0.)
    """

    def __init__(self, name=None):
        super(Counter, self).__init__(name=name)
        self.input.default.add_counts = 0
        self.output.n_counts = [0]

    def command(self, add_counts):
        if add_counts > 0:
            count = self.output.n_counts[-1] + add_counts
        else:
            count = self.output.n_counts[-1] + 1
        return {
            'n_counts': count
        }


class Compute(PrimitiveVertex):
    """
    """

    def __init__(self, name=None):
        super(Compute, self).__init__(name=name)

    def command(self, *args, **kwargs):
        function_ = self.input.function
        return function_(*args, **kwargs)


class Zeros(PrimitiveVertex):
    """
    A numpy vector of zeros that is pointer-compatible.
    Input attributes:
        shape (int/tuple): The shape of the array.
    Output attributes:
        zeros (numpy.ndarray): An array of numpy.float64 zeros.
    """

    def command(self, shape):
        return {
            'zeros': np.zeros(shape)
        }


class DeleteAtom(PrimitiveVertex):
    """
    Given a structure, deletes one of the atoms.
    Input attributes:
        structure (Atoms): The structure to delete an atom of.
        atom_id (int): Which atom to delete. (Default is 0, the 0th atom.)
    Output attributes:
        structure (Atoms): The new, modified structure.
        mask (numpy.ndarray): The integer ids shared by both the old and new structure.
    """

    def __init__(self, name=None):
        super(DeleteAtom, self).__init__(name=name)
        self.input.default.atom_id = 0

    def command(self, structure, atom_id):
        vacancy_structure = structure.copy()
        vacancy_structure.pop(atom_id)
        mask = np.delete(np.arange(len(structure)).astype(int), atom_id)

        return {
            'structure': vacancy_structure,
            'mask': mask
        }


class ExternalHamiltonian(PrimitiveVertex):
    """
    Manages calls to an external interpreter (e.g. Lammps, Vasp, Sphinx...) to produce energies, forces,
        and possibly other properties. The collected output can be expanded beyond forces and energies
        (e.g. to magnetic properties or whatever else the interpreting code produces) by modifying the
        `interesting_keys` in the input. The property must have a corresponding interactive getter for
        this property.
    Input attributes:
        ref_job_full_path (string): The full path to the hdf5 file of the job to use as a reference template.
            (Default is None.)
        project_path (string): The path of the project. To be specified ONLY when the jobs are already
            initialized outside of this vertex. If `project_path` is specified, `ref_job_full_path` must
            be set to None. (Default is None.)
        job_name (string): The name of the job. This is specified only when the jobs are already
            initialized outside of this vertex. (Default is None.)
        structure (Atoms): The structure for initializing the external Hamiltonian. Overwrites the reference
            job structure when provided. (Default is None, the reference job needs to have its structure set.)
        positions (numpy.ndarray): New positions to evaluate. Shape must match the shape of the structure.
            (Default is None, only necessary if positions are being updated.)
        cell (numpy.ndarray): The cell, if not same as that in the specified structure. (Default is None,
            same cell as in the structure.)
        interesting_keys (list[str]): String codes for output properties of the underlying job to collect.
            (Default is ['positions', 'forces', 'energy_pot', 'pressures', 'volume', 'cell'].)
    Output attributes:
        keys (list/numpy.ndarray): The output corresponding to the interesting keys.
    """

    def __init__(self, name=None):
        super(ExternalHamiltonian, self).__init__(name=name)
        self._fast_lammps_mode = True  # Set to false only to intentionally be slow for comparison purposes
        self._job_project_path = None
        self._job = None
        self._job_name = None
        id_ = self.input.default
        id_.ref_job_full_path = None
        id_.project_path = None
        id_.job_name = None
        id_.structure = None
        id_.positions = None
        id_.cell = None
        id_.interesting_keys = ['positions', 'forces', 'energy_pot', 'pressures', 'volume', 'cell']

    def command(self, ref_job_full_path, project_path, job_name, structure, positions, cell, interesting_keys):

        if self._job_project_path is None:
            if project_path is None and ref_job_full_path is not None:
                self._job_project_path, self._job_name = self._initialize(self.get_graph_location(),
                                                                          ref_job_full_path,
                                                                          structure,
                                                                          self._fast_lammps_mode
                                                                          )
            elif project_path is not None and ref_job_full_path is None:
                self._job_project_path = project_path
                self._job_name = job_name
            else:
                raise AttributeError('Please specify valid project path OR ref_job_full_path, but not both!')

        if self._job is None:
            self._reload()
        elif not self._job.interactive_is_activated():
            self._job.status.running = True
            self._job.interactive_open()
            self._job.interactive_initialize_interface()

        if isinstance(self._job, LammpsInteractive) and self._fast_lammps_mode:
            if positions is not None:
                self._job.interactive_positions_setter(positions)
            if cell is not None:
                self._job.interactive_cells_setter(cell)
            self._job._interactive_lib_command(self._job._interactive_run_command)
        elif isinstance(self._job, GenericInteractive):
            # DFT codes are slow enough that we can run them the regular way and not care
            # Also we might intentionally run Lammps slowly for comparison purposes
            if positions is not None:
                self._job.structure.positions = positions
            if cell is not None:
                self._job.structure.cell = cell
            self._job.calc_static()
            self._job.run()
        else:
            raise TypeError('Job of class {} is not compatible.'.format(self._job.__class__))

        return {key: self._get_interactive_value(key) for key in interesting_keys}

    @staticmethod
    def _initialize(graph_location, ref_job_full_path, structure, fast_lammps_mode, name=None):
        """
        Initialize / create the interactive job and save it.
        """
        if name is None:
            name = graph_location + '_job'
        project_path, ref_job_path = split(ref_job_full_path)
        pr = Project(path=project_path)
        ref_job = pr.load(ref_job_path)
        job = ref_job.copy_to(
            project=pr,
            new_job_name=name,
            input_only=True,
            new_database_entry=True
        )
        if structure is not None:
            job.structure = structure
        if isinstance(job, GenericInteractive):
            job.interactive_open()
            if isinstance(job, LammpsInteractive) and fast_lammps_mode:
                # Note: This might be done by default at some point in LammpsInteractive,
                # and could then be removed here
                job.interactive_flush_frequency = 10 ** 10
                job.interactive_write_frequency = 10 ** 10
            job.calc_static()
            job.run()
        else:
            raise TypeError('Job of class {} is not compatible.'.format(ref_job.__class__))

        return job.project.path, job.job_name

    def _reload(self):
        """
        Reload a saved job from its `project_path` and `job_name`.
        """
        pr = Project(path=self._job_project_path)
        self._job = pr.load(self._job_name)
        self._job.interactive_open()
        self._job.interactive_initialize_interface()

    def _get_interactive_value(self, key):
        """
        Get output corresponding to the `interesting_keys` from interactive job.
        """
        if key == 'positions':
            val = np.array(self._job.interactive_positions_getter())
        elif key == 'forces':
            val = np.array(self._job.interactive_forces_getter())
        elif key == 'energy_pot':
            val = self._job.interactive_energy_pot_getter()
        elif key == 'pressures':
            val = np.array(self._job.interactive_pressures_getter())
        elif key == 'volume':
            val = self._job.interactive_volume_getter()
        elif key == 'cell':
            val = np.array(self._job.interactive_cells_getter())
        else:
            raise NotImplementedError
        return val

    def finish(self):
        """
        Close the interactive job.
        """
        super(ExternalHamiltonian, self).finish()
        if self._job is not None:
            self._job.interactive_close()


class CreateJob(ExternalHamiltonian):
    """
    Creates a job of an external interpreter (e.g. Lammps, Vasp, Sphinx...) outside of the `ExternalHamiltonian`.
        This vertex does not run the interpreter, but only creates the job and saves it.
    Input attributes:
        ref_job_full_path (string): The full path to the hdf5 file of the job to use as a reference template.
            (Default is None.)
        n_images (int): Number of jobs to create. (Default is 5.)
        structure (Atoms): The structure for initializing the external Hamiltonian. Overwrites the reference
            job structure when provided. (Default is None, the reference job needs to have its structure set.)
    Output attributes:
        project_path (list/string): The path of the project.
        job_names (list/string): The name of the job.
    """

    def __init__(self, name=None):
        super(CreateJob, self).__init__(name=name)
        self._fast_lammps_mode = True
        self._project_path = None
        self._job_names = None
        id_ = self.input.default
        id_.ref_job_full_path = None
        id_.n_images = 5
        id_.structure = None

    def command(self, ref_job_full_path, n_images, structure, *args, **kwargs):
        graph_location = self.get_graph_location()
        job_names = []
        project_path = []
        for i in np.arange(n_images):
            name = graph_location + '_' + str(i)
            output = self._initialize(graph_location, ref_job_full_path, structure,
                                      self._fast_lammps_mode, name)
            project_path.append(output[0])
            job_names.append(output[1])
        self._project_path = project_path
        self._job_names = job_names

        return {
            'project_path': project_path,
            'job_names': job_names
        }

    def finish(self):
        """
        Close the interactive job.
        """
        super(CreateJob, self).finish()
        if all(v is not None for v in [self._project_path, self._job_names]):
            pr = Project(path=self._project_path[-1])
            for jn in self._job_names:
                job = pr.load(jn)
                if isinstance(job, GenericInteractive):
                    job.interactive_close()
                    job.status.finished = True


class MinimizeReferenceJob(ExternalHamiltonian):
    """
    Minimizes a job using an external interpreter (e.g. Lammps, Vasp, Sphinx...) outside of the `ExternalHamiltonian`.
        This vertex minimizes the job at constant volume or constant pressure.
    Input attributes:
        ref_job_full_path (string): The full path to the hdf5 file of the job to use as a reference template.
            (Default is None.)
        pressure (float): Pressure of the system. (Default is None.)
        structure (Atoms): The structure for initializing the external Hamiltonian. Overwrites the reference
            job structure when provided. (Default is None, the reference job needs to have its structure set.)
    Output attributes:
        energy_pot (float): The evaluated potential energy.
        forces (numpy.ndarray): The evaluated forces.
    """

    def __init__(self, name=None):
        super(MinimizeReferenceJob, self).__init__(name=name)
        self._fast_lammps_mode = True
        id_ = self.input.default
        id_.ref_job_full_path = None
        id_.pressure = None
        id_.structure = None

    def command(self, ref_job_full_path, structure, pressure=None, *args, **kwargs):
        graph_location = self.get_graph_location()
        name = 'minimize_ref_job'
        project_path, job_name = self._initialize(graph_location, ref_job_full_path, structure,
                                                  self._fast_lammps_mode, name)
        pr = Project(path=project_path)
        job = pr.load(job_name)
        job.structure = structure
        job.calc_minimize(pressure=pressure)
        job.run(delete_existing_job=True)
        if isinstance(job, GenericInteractive):
            job.interactive_close()

        return {
            'energy_pot': job.output.energy_pot[-1],
            'forces': job.output.forces[-1]
        }


class RemoveJob(PrimitiveVertex):
    """
    Remove an existing job/s from a project path.
    Input attributes:
        project_path (string): The path of the project. (Default is None.)
        job_names (string): The names of the jobs to be removed. (Default is None.)
    """
    def __init__(self, name=None):
        super(RemoveJob, self).__init__(name=name)
        self.input.default.project_path = None
        self.input.default.job_names = None

    def command(self, project_path, job_names):
        if all(v is not None for v in [project_path, job_names]):
            pr = Project(path=project_path)
            for name in job_names:
                pr.remove_job(name)


class GradientDescent(PrimitiveVertex):
    """
    Simple gradient descent update for positions in `flex_output` and structure.
    Input attributes:
        gamma0 (float): Initial step size as a multiple of the force. (Default is 0.1.)
        fix_com (bool): Whether the center of mass motion should be subtracted off of the position update.
            (Default is True)
        use_adagrad (bool): Whether to have the step size decay according to adagrad. (Default is False)
        output_displacements (bool): Whether to return the per-atom displacement vector in the output dictionary.
    Output attributes:
        positions (numpy.ndarray): The updated positions.
        displacements (numpy.ndarray): The displacements, if `output_displacements` is True.
    TODO: Fix adagrad bug when GradientDescent is passed as a Serial vertex
    """

    def __init__(self, name=None):
        super(GradientDescent, self).__init__(name=name)
        self._accumulated_force = 0
        id_ = self.input.default
        id_.gamma0 = 0.1
        id_.fix_com = True
        id_.use_adagrad = False

    def command(self, positions, forces, gamma0, use_adagrad, fix_com, mask=None, masses=None,
                output_displacements=True):

        positions = np.array(positions)
        forces = np.array(forces)
        unmasked_positions = None

        if mask is not None:
            masked = True
            mask = np.array(mask)
            # Preserve data
            unmasked_positions = positions.copy()

            # Mask input data
            positions = positions[mask]
            forces = forces[mask]
            masses = np.array(masses)[mask]
        else:
            masked = False

        if use_adagrad:
            self._accumulated_force += np.sqrt(np.sum(forces * forces))
            gamma0 /= self._accumulated_force

        pos_change = gamma0 * np.array(forces)

        if fix_com:
            masses = np.array(masses)[:, np.newaxis]
            total_mass = np.sum(masses)
            com_change = np.sum(pos_change * masses, axis=0) / total_mass
            pos_change -= com_change
        # TODO: fix angular momentum

        new_pos = positions + pos_change

        if masked:
            unmasked_positions[mask] = new_pos
            new_pos = unmasked_positions
            disp = np.zeros(unmasked_positions.shape)
            disp[mask] = pos_change
        else:
            disp = pos_change

        if output_displacements:
            return {
                'positions': new_pos,
                'displacements': disp
            }
        else:
            return {
                'positions': new_pos
            }


class InitialPositions(PrimitiveVertex):
    """
    Assigns initial positions. If no initial positions are specified, interpolates between the positions of the
        initial and the final structures.
    Input attributes:
        initial_positions (list/numpy.ndarray): The initial positions (Default is None)
        structure_initial (Atoms): The initial structure
        structure_final (Atoms): The final structure
        n_images (int): Number of structures to interpolate
    Output attributes:
        initial_positions (list/numpy.ndarray): if initial_positions is None, a list of (n_images) positions
            interpolated between the positions of the initial and final structures. Else, initial_positions
    """

    def __init__(self, name=None):
        super(InitialPositions, self).__init__(name=name)
        self.input.default.initial_positions = None

    def command(self, initial_positions, structure_initial, structure_final, n_images):

        if initial_positions is None:
            pos_i = structure_initial.positions
            pos_f = structure_final.positions
            displacement = structure_initial.find_mic(pos_f - pos_i)

            initial_positions = []
            for n, mix in enumerate(np.linspace(0, 1, n_images)):
                initial_positions.append(pos_i + (mix * displacement))

        else:
            if len(initial_positions) != n_images:
                raise TypeError("Length of positions is not the same as n_images!")

        return {
            'initial_positions': initial_positions
        }


class HarmonicHamiltonian(PrimitiveVertex):
    """
    Treat the atoms in the structure as harmonic oscillators and calculate the forces on each atom, and the total
        potential energy of the structure. If the spring constant is specified, the atoms act as Einstein atoms
        (independent of each other). If the Hessian / force constant matrix is specified, the atoms act as Debye
        atoms.
    Input attributes:
        positions (numpy.ndarray): Current positions of the atoms.
        reference_positions (numpy.ndarray): Equilibrium positions of the atoms.
        spring_constant (float): A single spring / force constant that is used to compute the restoring forces
            on each atom. (Default is 1.)
        force_constants (NxNx3x3 matrix): The Hessian matrix, obtained from, for ex. Phonopy. (Default is None, treat
            the atoms as independent harmonic oscillators (Einstein atoms).
        structure (Atoms): The reference structure.
        mask (numpy.array): Which of the atoms to consider. The other atoms are ignored. (Default is None, consider
            all atoms.)
        eq_energy (float): The minimized potential energy of the static (expanded) structure. (Default is None.)
    Output attributes:
        energy_pot (float): The harmonic potential energy.
        forces (numpy.ndarray): The harmonic forces.
    """

    def __init__(self, name=None):
        super(HarmonicHamiltonian, self).__init__(name=name)
        id_ = self.input.default
        id_.spring_constant = None
        id_.force_constants = None

    def command(self, positions, reference_positions, structure, spring_constant=None, force_constants=None,
                mask=None, eq_energy=None):

        dr = structure.find_mic(positions - reference_positions)
        if spring_constant is not None and force_constants is None:
            if mask is not None:
                forces = -spring_constant * dr[mask]
                energy = -0.5 * np.dot(dr[mask], forces)
            else:
                forces = -spring_constant * dr
                energy = -0.5 * np.tensordot(dr, forces)

        elif force_constants is not None and spring_constant is None:
            transformed_force_constants = self.transform_force_constants(force_constants, len(structure.positions))
            transformed_displacements = self.transform_displacements(dr)
            transformed_forces = -np.dot(transformed_force_constants, transformed_displacements)
            retransformed_forces = self.retransform_forces(transformed_forces, dr)
            if mask is not None:
                forces = retransformed_forces[mask]
                energy = -0.5 * np.dot(forces, dr[mask])
            else:
                forces = retransformed_forces
                energy = -0.5 * np.tensordot(forces, dr)

        else:
            raise TypeError('Please specify either a spring constant or the force constant matrix')

        if eq_energy is not None:
            energy += eq_energy

        return {
            'forces': forces,
            'energy_pot': energy
        }

    @staticmethod
    def transform_force_constants(force_constants, n_atoms):
        force_shape = np.shape(force_constants)
        if force_shape[2] == 3 and force_shape[3] == 3:
            force_reshape = force_shape[0] * force_shape[2]
            transformed_force_constants = np.transpose(
                force_constants,
                (0, 2, 1, 3)
            ).reshape((force_reshape, force_reshape))
        elif force_shape[1] == 3 and force_shape[3] == 3:
            transformed_force_constants = np.array(force_constants).reshape(3 * n_atoms, 3 * n_atoms)

        return transformed_force_constants

    @staticmethod
    def transform_displacements(displacements):
        return displacements.reshape(displacements.shape[0] * displacements.shape[1])

    @staticmethod
    def retransform_forces(transformed_forces, displacements):
        return transformed_forces.reshape(displacements.shape)


class LangevinThermostat(PrimitiveVertex):
    """
    Calculates the necessary forces for a Langevin thermostat based on drag and a random kick.
    Input dictionary:
        velocities (numpy.ndarray): The per-atom velocities in angstroms/fs.
        masses (numpy.ndarray): The per-atom masses in atomic mass units.
        temperature (float): The target temperature. (Default is 0 K, which just applies a drag force.)
        damping_timescale (float): Damping timescale. (Default is 100 fs.)
        time_step (float): MD time step in fs. (Default is 1 fs.)
        fix_com (bool): Whether to ensure the net force is zero. (Default is True.)
    Output dictionary:
        forces (numpy.ndarray): The per-atom forces from the thermostat.
    TODO: Make a version that uses uniform random numbers like Lammps does (for speed)
    """

    def __init__(self, name=None):
        super(LangevinThermostat, self).__init__(name=name)
        id_ = self.input.default
        id_.temperature = 0.
        id_.damping_timescale = 100.
        id_.time_step = 1.
        id_.fix_com = True

    def command(self, velocities, masses, temperature, damping_timescale, time_step, fix_com):
        # Ensure that masses are a commensurate shape
        masses = np.array(masses)[:, np.newaxis]
        gamma = masses / damping_timescale
        np.random.seed()
        noise = np.sqrt(2 * (gamma / time_step) * KB * temperature) * np.random.randn(*velocities.shape)
        drag = -gamma * velocities
        thermostat_forces = noise + drag

        if fix_com:  # Apply zero net force
            thermostat_forces -= np.mean(noise, axis=0)

        return {
            'forces': thermostat_forces
        }


class Max(PrimitiveVertex):
    """
    Numpy's amax (except no `out` option). Docstrings directly copied from the `numpy docs`_. The `initial`
    field is renamed `initial_val` to not conflict with the input dictionary.
    _`numpy docs`: https://docs.scipy.org/doc/numpy/reference/generated/numpy.amax.html
    Input attributes:
        a (array_like): Input data.
        axis (None/int/tuple of ints): Axis or axes along which to operate. By default, flattened input is used. If
            this is a tuple of ints, the maximum is selected over multiple axes, instead of a single axis or all the
            axes as before.
        keepdims (bool): If this is set to True, the axes which are reduced are left in the result as dimensions with
            size one. With this option, the result will broadcast correctly against the input array. If the default
            value is passed, then *keepdims* will not be passed through to the `amax` method of sub-classes of
            `ndarray`, however any non-default value will be. If the sub-class’ method does not implement *keepdims*
            any exceptions will be raised.
    Output attributes:
        amax (numpy.ndarray/scalar): Maximum of *a*. If axis is None, the result is a scalar value. If axis is given,
            the result is an array of dimension `a.ndim - 1`.
    Note: This misses the new argument `initial` in the latest version of Numpy.
    """

    def __init__(self, name=None):
        super(Max, self).__init__(name=name)
        id_ = self.input.default
        id_.axis = None
        id_.keepdims = False
        id_.initial_val = None

    def command(self, a, axis, keepdims, initial_val):
        return {
            'amax': np.amax(a, axis=axis, keepdims=keepdims)
        }


class NEBForces(PrimitiveVertex):
    """
    Given a list of positions, forces, and energies for each image along some transition, calculates the tangent
        direction along the transition path for each image and returns new forces which have their original value
        perpendicular to this tangent, and 'spring' forces parallel to the tangent. Endpoints forces are set to
        zero, so start with relaxed endpoint structures.
    Note: All images must use the same cell and contain the same number of atoms.
    Input attributes:
        positions (list/numpy.ndarray): The positions of the images. Each element should have the same shape, and
            the order of the images is relevant!
        energies (list/numpy.ndarray): The potential energies associated with each set of positions. (Not always
            needed.)
        forces (list/numpy.ndarray): The forces from the underlying energy landscape on which the NEB is being run.
            Must have the same shape as `positions`. (Not always needed.)
        structure (Atoms): The reference structure.
        spring_constant (float): Spring force between atoms in adjacent images. (Default is 1.0 eV/angstrom^2.)
        tangent_style ('plain'/'improved'/'upwinding'): How to calculate the image tangent in 3N-dimensional space.
            (Default is 'upwinding', which requires energies to be set.)
        use_climbing_image (bool): Whether to replace the force with one that climbs along the tangent direction for
            the job with the highest energy. (Default is True, which requires energies and a list of forces to
            be set.)
        smoothing (float): Strength of the smoothing spring when consecutive images form an angle. (Default is None,
            do not apply such a force.)
    Output attributes:
        forces (list): The forces after applying nudged elastic band method.
    TODO:
        Add references to papers (Jonsson and others)
        Implement Sheppard's equations for variable cell shape
    """

    def __init__(self, name=None):
        super(NEBForces, self).__init__(name=name)
        id_ = self.input.default
        id_.spring_constant = 1.
        id_.tangent_style = 'upwinding'
        id_.use_climbing_image = True
        id_.smoothing = None

    def command(self, positions, energies, forces, structure, spring_constant, tangent_style, smoothing,
                use_climbing_image):

        if use_climbing_image:
            climbing_image_index = np.argmax(energies)
        else:
            climbing_image_index = None

        n_images = len(positions)
        neb_forces = []
        for i, pos in enumerate(positions):
            if i == 0 or i == n_images - 1:
                neb_forces.append(np.zeros(pos.shape))
                continue

            # Otherwise calculate the spring forces
            pos_left = positions[i - 1]
            pos_right = positions[i + 1]

            # Get displacement to adjacent images
            dr_left = structure.find_mic(pos - pos_left)
            dr_right = structure.find_mic(pos_right - pos)

            # Get unit vectors to adjacent images
            tau_left = dr_left / np.linalg.norm(dr_left)
            tau_right = dr_right / np.linalg.norm(dr_right)

            # Calculate the NEB tangent vector
            if tangent_style == 'plain':
                tau = self.normalize(dr_right + dr_left)
            elif tangent_style == 'improved':
                tau = self.normalize(tau_left + tau_right)
            elif tangent_style == 'upwinding':
                en_left = energies[i - 1]
                en = energies[i]
                en_right = energies[i + 1]
                tau = self._get_upwinding_tau(tau_left, tau_right, en_left, en, en_right)
            else:
                # This branch should be inaccessible due to the setter...
                raise KeyError("No such tangent_style: " + str(tangent_style))

            if smoothing is not None:
                force_smoothing = (self.saturating_angle_control(dr_left, dr_right) *
                                   smoothing * (dr_right - dr_left))
            else:
                force_smoothing = 0

            # Decompose the original forces
            input_force = forces[i]
            input_force_parallel = np.sum(input_force * tau) * tau
            input_force_perpendicular = input_force - input_force_parallel

            # If climbing image is activated, push the highest energy image up along tau while
            # relaxing otherwise
            if use_climbing_image and i == climbing_image_index:
                neb_forces.append(input_force_perpendicular - input_force_parallel)
            else:
                dr_mag = np.linalg.norm(dr_right) - np.linalg.norm(dr_left)
                force_spring = spring_constant * dr_mag * tau
                neb_forces.append(input_force_perpendicular + force_spring + force_smoothing)

        return {
            'forces': neb_forces
        }

    def _get_upwinding_tau(self, tau_left, tau_right, en_left, en, en_right):
        """
        Take direction to the higher-energy image.
        """
        # Find the relative energies
        den_left = en - en_left
        den_right = en - en_right

        # Create booleans to describe the energetic position of the job
        high_left = en_left > en_right
        high_right = en_right > en_left
        extrema = (en_right > en < en_left) or (en_right < en > en_left)

        den_mag_min = min(abs(den_left), abs(den_right))
        den_mag_max = max(abs(den_left), abs(den_right))
        # Calculate the NEB direction
        tau = tau_left * (den_mag_max * high_left + extrema * den_mag_min * high_right) + \
              tau_right * (den_mag_max * high_right + extrema * den_mag_min * high_left)
        return self.normalize(tau)

    @staticmethod
    def normalize(vec):
        l1 = np.linalg.norm(vec)
        if l1 == 0:
            return vec
        return vec / l1

    @staticmethod
    def saturating_angle_control(dr_left, dr_right):
        """
        Return 0 when displacement to left and right neighbours is parallel, and run smoothly to 1
        then saturate at this value when these displacement vectors form an acute angle.
        """
        dr_left_mag = np.linalg.norm(dr_left)
        dr_right_mag = np.linalg.norm(dr_right)
        cos_angle = np.sum(dr_left * dr_right) / (dr_left_mag * dr_right_mag)
        return 0.5 * (1 + np.cos(np.pi * cos_angle))


class Norm(PrimitiveVertex):
    """
    Numpy's linalg norm. Docstrings directly copied from the `numpy docs`_.
    _`numpy docs`: https://docs.scipy.org/doc/numpy/reference/generated/numpy.linalg.norm.html
    Input attributes:
        x (array_like): Input array. If axis is None, x must be 1-D or 2-D.
        ord (non-zero int/inf/-inf/'fro'/'nuc'): Order of the norm. inf means numpy’s inf object. (Default is 2).
        axis (int/2-tuple of ints/None): If axis is an integer, it specifies the axis of x along which to compute the
            vector norms. If axis is a 2-tuple, it specifies the axes that hold 2-D matrices, and the matrix norms of
            these matrices are computed. If axis is None then either a vector norm (when x is 1-D) or a matrix norm
            (when x is 2-D) is returned. (Default is None.)
        keepdims (bool): If this is set to True, the axes which are normed over are left in the result as dimensions
            with size one. With this option the result will broadcast correctly against the original x. (Default is
            False)
    Output attributes:
        n (float/numpy.ndarray): Norm of the matrix or vector(s).
    """

    def __init__(self, name=None):
        super(Norm, self).__init__(name=name)
        id_ = self.input.default
        id_.ord = 2
        id_.axis = None
        id_.keepdims = False

    def command(self, x, ord, axis, keepdims):
        return {
            'n': np.linalg.norm(x, ord=ord, axis=axis, keepdims=keepdims)
        }


class Overwrite(PrimitiveVertex):
    """
    Overwrite particular entries of an array with new values.
    Input attributes:
        target (numpy.ndarray): The target array.
        mask (numpy.array): The indices of the target that will be replaced.
        new_values (numpy.ndarray): The targeted indices will be replaced by these values.
    Output attributes:
        overwritten (numpy.ndarray): The overwritten array.
    """

    def command(self, target, mask, new_values):
        overwritten = np.array(target)
        overwritten[mask] = new_values
        return {
            'overwritten': overwritten
        }


class RandomVelocity(PrimitiveVertex):
    """
    Generates a set of random velocities which (on average) have give the requested temperature.
        Hard-coded for 3D systems.
    Input attributes:
        temperature (float): The temperature of the velocities (in Kelvin).
        masses (numpy.ndarray/list): The masses of the atoms.
        overheat_fraction (float): The fraction by which to overheat the initial velocities. This can be useful for
            more quickly equilibrating a system whose initial structure is its fully relaxed positions -- in which
            case equipartition of energy tells us that the kinetic energy should be initialized to double the
            desired value. (Default is 2.0, assume energy equipartition is a good idea.)
    Output attributes:
        velocities (numpy.ndarray): Per-atom velocities.
        energy_kin (float): Total kinetic energy of all atoms.
        n_atoms (int): Number of atoms.
    """

    def __init__(self, name=None):
        super(RandomVelocity, self).__init__(name=name)
        self.input.default.overheat_fraction = 2.

    def command(self, temperature, masses, overheat_fraction):
        masses = np.array(masses)[:, np.newaxis]
        vel_scale = np.sqrt(EV_TO_U_ANGSQ_PER_FSSQ * KB * temperature / masses) * np.sqrt(overheat_fraction)
        np.random.seed(0)
        vel_dir = np.random.randn(len(masses), 3)
        vel = vel_scale * vel_dir
        vel -= np.mean(vel, axis=0)
        energy_kin = 0.5 * np.sum(masses * vel * vel) * U_ANGSQ_PER_FSSQ_TO_EV
        return {
            'velocities': vel,
            'energy_kin': energy_kin,
            'n_atoms': len(vel)
        }


class SphereReflection(PrimitiveVertex):
    """
    Checks whether each atom in a structure is within a cutoff radius of its reference position; if not, reverts
        the positions and velocities of *all atoms* to an earlier state and reverses those earlier velocities.
        Forces from the current and previous time step can be provided optionally.
    Input attributes:
        reference_positions (numpy.ndarray): The reference positions to check the distances from.
        positions (numpy.ndarray): The positions to check.
        velocities (numpy.ndarray): The velocities corresponding to the positions at this time.
        previous_positions (numpy.ndarray): The previous positions to revert to.
        previous_velocities (numpy.ndarray): The previous velocities to revert to and reverse.
        structure (Atoms): The reference structure.
        cutoff_distance (float): The cutoff distance from the reference position to trigger reflection.
        use_reflection (boolean): Turn on or off `SphereReflection`
        total_steps (int): The total number of times `SphereReflection` is called so far.
    Output attributes:
        positions (numpy.ndarray): The (possibly reverted) positions.
        velocities (numpy.ndarray): The (possibly reverted and reversed) velocities.
        reflected_mask (bool): Whether or not reflection occurred.
        total_steps (int): The total number of times `SphereReflection` is called.
    """

    def __init__(self, name=None):
        super(SphereReflection, self).__init__(name=name)
        id_ = self.input.default
        id_.use_reflection = True
        id_.total_steps = 0

    def command(self, reference_positions, cutoff_distance, positions, velocities, previous_positions,
                previous_velocities, structure, use_reflection, total_steps):
        total_steps += 1
        distance = np.linalg.norm(structure.find_mic(reference_positions - positions), axis=-1)
        is_at_home = (distance < cutoff_distance)[:, np.newaxis]
        if np.all(is_at_home) or use_reflection is False:
            return {
                'positions': positions,
                'velocities': velocities,
                'reflected': False,
                'total_steps': total_steps
            }
        else:
            return {
                'positions': previous_positions,
                'velocities': -previous_velocities,
                'reflected': True,
                'total_steps': total_steps
            }


class SphereReflectionPerAtom(PrimitiveVertex):
    """
    Checks whether each atom in a structure is within a cutoff radius of its reference position; if not, reverts
        the positions and velocities of *the violating atoms* to an earlier state and reverses those earlier
        velocities. The remaining atoms are unaffected.
    Input attributes:
        reference_positions (numpy.ndarray): The reference positions to check the distances from.
        positions (numpy.ndarray): The positions to check.
        velocities (numpy.ndarray): The velocities corresponding to the positions at this time.
        previous_positions (numpy.ndarray): The previous positions to revert to.
        previous_velocities (numpy.ndarray): The previous velocities to revert to and reverse.
        structure (Atoms): The reference structure.
        cutoff_distance (float): The cutoff distance from the reference position to trigger reflection.
        use_reflection (boolean): Turn on or off `SphereReflectionPerAtom`
        total_steps (int): The total number of times `SphereReflectionPerAtom` is called so far.
    Output attributes:
        positions (numpy.ndarray): The (possibly reverted) positions.
        velocities (numpy.ndarray): The (possibly reverted and reversed) velocities.
        reflected_mask (numpy.ndarray): A boolean mask that is true for each atom who was reflected.
        total_steps (int): The total number of times `SphereReflectionPerAtom` is called.
    """

    def __init__(self, name=None):
        super(SphereReflectionPerAtom, self).__init__(name=name)
        id_ = self.input.default
        id_.use_reflection = True
        id_.total_steps = 0

    def command(self, reference_positions, cutoff_distance, positions, velocities, previous_positions,
                previous_velocities, structure, use_reflection, total_steps):
        total_steps += 1
        if use_reflection:
            distance = np.linalg.norm(structure.find_mic(reference_positions - positions), axis=-1)
            is_at_home = (distance < cutoff_distance)[:, np.newaxis]
            is_away = 1 - is_at_home
        else:
            is_at_home = np.ones(len(reference_positions))
            is_away = 1 - is_at_home

        return {
            'positions': is_at_home * positions + is_away * previous_positions,
            'velocities': is_at_home * velocities + is_away * -previous_velocities,
            'reflected': is_away.astype(bool).flatten(),
            'total_steps': total_steps
        }


class CutoffDistance(PrimitiveVertex):
    """
    Compute the cutoff distance for SphereReflection.
    Input attributes:
        structure (Atoms): The reference structure.
        cutoff_factor (float): The cutoff is obtained by taking the first nearest neighbor distance and multiplying
            it by the cutoff factor. A default value of 0.4 is chosen, because taking a cutoff factor of ~0.5
            sometimes let certain reflections off the hook, and we do not want that to happen.
    Output attributes:
        cutoff_distance (float): The cutoff distance.
    """

    def command(self, structure, cutoff_factor=0.5):
        try:
            nn_list = structure.get_neighbors(num_neighbors=1)
        except ValueError:
            nn_list = structure.get_neighbors(num_neighbors=1)

        cutoff_distance = nn_list.distances[0] * cutoff_factor

        return {
            'cutoff_distance': cutoff_distance[-1]
        }


class Slice(PrimitiveVertex):
    """
    Slices off the masked positions from the input vector.
    Input attributes:
        vector (numpy.ndarray): The input vector.
        mask (numpy.ndarray): The integer ids shared by both the old and new structure.
        ensure_interable mask (bool):
    Output attributes:
        sliced (numpy.ndarray): The sliced vector.
    """

    def __init__(self, name=None):
        super(Slice, self).__init__(name=name)
        self.input.default.ensure_iterable_mask = False

    def command(self, vector, mask, ensure_iterable_mask):
        if ensure_iterable_mask:
            mask = ensure_iterable(mask)
        return {
            'sliced': vector[mask]
        }


class Transpose(PrimitiveVertex):
    """
    Transposes the input matrix.
    Input attributes:
        matrix (numpy.ndarray): The input matrix.
    Output attributes:
        matrix_transpose (numpy.ndarray): The transposed matrix.
    """

    def command(self, matrix):
        transpose = list(map(list, zip(*matrix)))
        return {
            'matrix_transpose': transpose
        }


class VerletParent(PrimitiveVertex, ABC):
    """
    A parent class for holding code which is shared between both position and velocity updates in two-step
        Velocity Verlet.
    Input attributes:
        time_step (float): MD time step in fs. (Default is 1 fs.)
        temperature (float): The target temperature. (Default is None, no thermostat is used.)
        damping_timescale (float): Damping timescale in fs. (Default is None, no thermostat is used.)
    TODO: VerletVelocityUpdate should *always* have its velocity input wired to the velocity outupt of
          VerletPositionUpdate. This implies to me that we need some structure *other* than two fully independent
          nodes. It would also be nice to syncronize, e.g. the thermostat and timestep input which is also the same
          for both. However, I haven't figured out how to do that in the confines of the current graph traversal
          and hdf5 setup.
    """

    def __init__(self, name=None):
        super(VerletParent, self).__init__(name=name)
        id_ = self.input.default
        id_.time_step = 1.
        id_.temperature = None
        id_.temperature_damping_timescale = None

    @abstractmethod
    def command(self, *arg, **kwargs):
        pass

    @staticmethod
    def reshape_masses(masses):
        if len(np.array(masses).shape) == 1:
            masses = np.array(masses)[:, np.newaxis]
        return masses

    @staticmethod
    def convert_to_acceleration(forces, masses):
        return forces * EV_TO_U_ANGSQ_PER_FSSQ / masses

    @staticmethod
    def langevin_delta_v(temperature, time_step, masses, damping_timescale, velocities):
        """
        Velocity changes due to the Langevin thermostat.
        Args:
            temperature (float): The target temperature in K.
            time_step (float): The MD time step in fs.
            masses (numpy.ndarray): Per-atom masses in u with a shape (N_atoms, 1).
            damping_timescale (float): The characteristic timescale of the thermostat in fs.
            velocities (numpy.ndarray): Per-atom velocities in angstrom/fs.
        Returns:
            (numpy.ndarray): Per atom accelerations to use for changing velocities.
        """
        drag = -0.5 * time_step * velocities / damping_timescale
        np.random.seed()
        noise = np.sqrt(EV_TO_U_ANGSQ_PER_FSSQ * KB * temperature * time_step / (masses * damping_timescale)) \
                * np.random.randn(*velocities.shape)
        noise -= np.mean(noise, axis=0)
        return drag + noise


class VerletPositionUpdate(VerletParent):
    """
    First half of Velocity Verlet, where positions are updated and velocities are set to their half-step value.
    Input attributes:
        positions (numpy.ndarray): The per-atom positions in angstroms.
        velocities (numpy.ndarray): The per-atom velocities in angstroms/fs.
        forces (numpy.ndarray): The per-atom forces in eV/angstroms.
        masses (numpy.ndarray): The per-atom masses in atomic mass units.
        time_step (float): MD time step in fs. (Default is 1 fs.)
        temperature (float): The target temperature. (Default is None, no thermostat is used.)
        damping_timescale (float): Damping timescale in fs. (Default is None, no thermostat is used.)
    Output attributes:
        positions (numpy.ndarray): The new positions on time step in the future.
        velocities (numpy.ndarray): The new velocities *half* a time step in the future.
    """

    def command(self, positions, velocities, forces, masses, time_step, temperature, temperature_damping_timescale):
        masses = self.reshape_masses(masses)
        acceleration = self.convert_to_acceleration(forces, masses)
        vel_half = velocities + 0.5 * acceleration * time_step
        if temperature_damping_timescale is not None:
            vel_half += self.langevin_delta_v(
                temperature,
                time_step,
                masses,
                temperature_damping_timescale,
                velocities
            )
        pos_step = positions + vel_half * time_step

        return {
            'positions': pos_step,
            'velocities': vel_half
        }


class VerletVelocityUpdate(VerletParent):
    """
    Second half of Velocity Verlet, where velocities are updated. Forces should be updated between the position
        and velocity updates
    Input attributes:
        velocities (numpy.ndarray): The per-atom velocities in angstroms/fs. These should be the half-step
            velocities output by `VerletPositionUpdate`.
        forces (numpy.ndarray): The per-atom forces in eV/angstroms. These should be updated since the last call
            of `VerletPositionUpdate`.
        masses (numpy.ndarray): The per-atom masses in atomic mass units.
        time_step (float): MD time step in fs. (Default is 1 fs.)
        temperature (float): The target temperature. (Default is None, no thermostat is used.)
        damping_timescale (float): Damping timescale in fs. (Default is None, no thermostat is used.)
    Output attributes:
        velocities (numpy.ndarray): The new velocities *half* a time step in the future.
        energy_kin (float): The total kinetic energy of the system in eV.
        instant_temperature (float): The instantaneous temperature, obtained from the total kinetic energy.
    """

    def command(self, velocities, forces, masses, time_step, temperature, temperature_damping_timescale):
        masses = self.reshape_masses(masses)
        acceleration = self.convert_to_acceleration(forces, masses)

        vel_step = velocities + 0.5 * acceleration * time_step
        if temperature_damping_timescale is not None:
            vel_step += self.langevin_delta_v(
                temperature,
                time_step,
                masses,
                temperature_damping_timescale,
                velocities
            )
        kinetic_energy = 0.5 * np.sum(masses * vel_step * vel_step) / EV_TO_U_ANGSQ_PER_FSSQ
        instant_temperature = (kinetic_energy * 2) / (3 * KB * len(velocities))

        return {
            'velocities': vel_step,
            'energy_kin': kinetic_energy,
            'instant_temperature': instant_temperature
        }


class VoronoiReflection(PrimitiveVertex):
    """
    Checks whether each atom in a structure is closest to its own reference site; if not, reverts the positions
        and velocities to an earlier state and reverses those earlier velocities.
    Input attributes:
        reference_positions (numpy.ndarray): The reference positions to check the distances from.
        positions (numpy.ndarray): The positions to check.
        velocities (numpy.ndarray): The velocities to check.
        previous_positions (numpy.ndarray): The previous positions to revert to.
        previous_velocities (numpy.ndarray): The previous velocities to revert to and reverse.
        pbc (numpy.ndarray/list): Three booleans declaring which dimensions have periodic boundary conditions for
            finding the minimum distance convention.
        cell (numpy.ndarray): The 3x3 cell vectors for pbcs.
    Output attributes:
        positions (numpy.ndarray): The (possibly reverted) positions.
        velocities (numpy.ndarray): The (possibly reverted and reversed) velocities.
        reflected_mask (numpy.ndarray): A boolean mask that is true for each atom who was reflected.
    WARNING: Outdated.
    """

    def __init__(self, name=None):
        super(VoronoiReflection, self).__init__(name=name)

    def command(self, reference_positions, positions, velocities, previous_positions, previous_velocities, pbc, cell):
        _, distance_matrix = get_distances(p1=reference_positions, p2=positions, cell=cell, pbc=pbc)
        closest_reference = np.argmin(distance_matrix, axis=0)
        is_at_home = (closest_reference == np.arange(len(positions)))[:, np.newaxis]
        is_away = 1 - is_at_home
        return {
            'positions': is_at_home * positions + is_away * previous_positions,
            'velocities': is_at_home * velocities + is_away * -previous_velocities,
            'reflected': is_away.astype(bool).flatten()
        }


class WeightedSum(PrimitiveVertex):
    """
    Given a list of vectors of with the same shape, calculates a weighted sum of the vectors. By default the
        weights are the inverse of the number of elements and the sum is just the mean.
    Input attributes:
        vectors (list/numpy.ndarray): The vectors to sum. (Masked) vectors must all be of the same length. If the
            the vectors are already in a numpy array, the 0th index should determine the vector.
        weights (list/numpy.ndarray): A 1D list of coefficients (floats) with the same length as `vectors`.
            (Default is None, which gives the simple mean.)
        masks (list): If not None, a mask must be passed for each vector to extract a sub-vector. The resulting
            collection of vectors and sub-vectors must all have the same length. If the mask for a given vector
            is `None`, all elements are retained. Otherwise the mask must be integer-like or boolean. (Default is
            None, do not mask any of the vectors.)
    Output attributes:
        weighted_sum (numpy.ndarray): The weighted sum, having the same shape as a (masked) input vector.
    """

    def __init__(self, name=None):
        super(WeightedSum, self).__init__(name=name)
        self.input.default.weights = None
        self.input.default.masks = None

    def command(self, vectors, weights, masks):
        # Handle weight defaults
        if weights is None:
            n = len(vectors)
            weights = np.ones() / n
        elif len(weights) != len(vectors):
            raise ValueError('The length of the weights and vectors must be comensurate, but were {} and {}'.format(
                len(weights), len(vectors)))

        # Mask vectors
        if masks is not None:
            for n, mask in enumerate(masks):
                if mask is not None:
                    if isinstance(mask[0], bool) or isinstance(mask[0], np.bool_):
                        mask = np.array(mask, dtype=bool)
                    else:
                        mask = np.array(mask, dtype=int)
                    vectors[n] = vectors[n][mask]

        # Prepare vectors for numpy operation
        vectors = np.array(vectors)

        # Dot vectors
        weighted_sum = np.tensordot(weights, vectors, axes=1)

        # If dot reduces to a single value, recast
        if len(weighted_sum.shape) == 0:
            weighted_sum = float(weighted_sum)

        return {
            'weighted_sum': weighted_sum
        }


class WelfordOnline(PrimitiveVertex):
    """
    Computes the cumulative mean and standard deviation.
    Note: The standard deviation calculated is for the population (ddof=0). For the sample (ddof=1) it would
        to be extended.
    Input attributes:
        sample (float/numpy.ndarray): The new sample. (Default is None.)
        mean (float/numpy.ndarray): The mean so far. (Default is None.)
        std (float/numpy.ndarray): The standard deviation so far. (Default is None.)
        n_samples (int): How many samples were used to calculate the existing `mean` and `std`.
    Output attributes:
        mean (float/numpy.ndarray): The new mean.
        std (float/numpy.ndarray): The new standard deviation.
        n_samples (int): The new number of samples.
    """

    def __init__(self, name=None):
        super(WelfordOnline, self).__init__(name=name)
        id_ = self.input.default
        id_.mean = None
        id_.std = None
        id_.n_samples = None

        op = Pointer(self.output)
        self.input.mean = op.mean[-1]
        self.input.std = op.std[-1]
        self.input.n_samples = op.n_samples[-1]

    def command(self, sample, mean, std, n_samples):
        if n_samples is None:
            new_mean = sample
            new_std = 0
            n_samples = 0
        else:
            new_mean, new_std = welford_online(sample, mean, std, n_samples)
        return {
            'mean': new_mean,
            'std': new_std,
            'n_samples': n_samples + 1
        }


class FEPExponential(PrimitiveVertex):
    """
    Compute the free energy perturbation exponential difference.
    Input attributes:
        u_diff (float): The energy difference between system B and system A.
        delta_lambda (float): The delta for the lambdas for the two systems A and B.
        temperature (float): The instantaneous temperature.
    Output attributes:
        exponential_difference (float): The exponential difference.
    """

    def command(self, u_diff, delta_lambda, temperature):
        return {
            'exponential_difference': np.exp(-u_diff * delta_lambda / (KB * temperature))
        }


class TILDPostProcess(PrimitiveVertex):
    """
    Post processing for the Harmonic and Vacancy TILD protocols, to remove the necessity to load interactive
        jobs after they have been closed, once the protocol has been executed.
    Input attributes:
        lambda_pairs (numpy.ndarray): The (`n_lambdas`, 2)-shaped array of mixing pairs.
        tild_mean (list): The mean of the computed integration points.
        tild_std (list): The standard deviation of the computed integration points.
        fep_exp_mean (list): The mean of the free energy perturbation exponential differences.
        fep_exp_mean (list): The standard deviation of the free energy perturbation exponential differences.
        temperature (float): The simulated temperature in K.
        n_samples (int): Number of samples used to calculate the means.
    Output attributes:
        tild_free_energy_mean (float): The mean calculated via thermodynamic integration.
        tild_free_energy_std (float): The standard deviation calculated via thermodynamic integration.
        tild_free_energy_se (float): The standard error calculated via thermodynamic integration.
        fep_free_energy_mean (float): The mean calculated via free energy perturbation.
        fep_free_energy_std (float): The standard deviation calculated via free energy perturbation.
        fep_free_energy_se (float): The standard error calculated via free energy perturbation.
    """

    def command(self, lambda_pairs, tild_mean, tild_std, fep_exp_mean, fep_exp_std, temperature, n_samples):

        tild_fe_mean, tild_fe_std, tild_fe_se = self.get_tild_free_energy(lambda_pairs, tild_mean, tild_std,
                                                                          n_samples)
        fep_fe_mean, fep_fe_std, fep_fe_se = self.get_fep_free_energy(fep_exp_mean, fep_exp_std, n_samples,
                                                                      temperature)
        return {
            'tild_free_energy_mean': tild_fe_mean,
            'tild_free_energy_std': tild_fe_std,
            'tild_free_energy_se': tild_fe_se,
            'fep_free_energy_mean': fep_fe_mean,
            'fep_free_energy_std': fep_fe_std,
            'fep_free_energy_se': fep_fe_se

        }

    @staticmethod
    def get_tild_free_energy(lambda_pairs, tild_mean, tild_std, n_samples):
        y = unumpy.uarray(tild_mean, tild_std)
        integral = simps(x=lambda_pairs[:, 0], y=y)
        mean = unumpy.nominal_values(integral)
        std = unumpy.std_devs(integral)
        tild_se = tild_std / np.sqrt(n_samples)
        y_se = unumpy.uarray(tild_mean, tild_se)
        integral_se = simps(x=lambda_pairs[:, 0], y=y_se)
        se = unumpy.std_devs(integral_se)

        return float(mean), float(std), float(se)

    @staticmethod
    def get_fep_free_energy(fep_exp_mean, fep_exp_std, n_samples, temperature):
        fep_exp_se = fep_exp_std / np.sqrt(n_samples)
        y = unumpy.uarray(fep_exp_mean, fep_exp_std)
        y_se = unumpy.uarray(fep_exp_mean, fep_exp_se)
        free_energy = 0
        free_energy_se = 0
        for (val, val_se) in zip(y, y_se):
            free_energy += -KB * temperature * unumpy.log(val)
            free_energy_se += -KB * temperature * unumpy.log(val_se)
        mean = unumpy.nominal_values(free_energy)
        std = unumpy.std_devs(free_energy)
        se = unumpy.std_devs(free_energy_se)

        return float(mean), float(std), float(se)


class BerendsenBarostat(PrimitiveVertex):
    """
    The Berendsen barostat which can be used for pressure control as elaborated in
    https://doi.org/10.1063/1.448118
    NOTE: Always use in conjunction with a thermostat, otherwise time integration will not be performed.
    The barostat only modifies the cell of the input structure, and scales the positions. Positions
    and velocities will only be updated by the thermostat.
    Input attributes:
        pressure (float): The pressure in GPa to be simulated (Default is None GPa)
        temperature (float): The temperature in K (Default is 0. K)
        box_pressures (numpy.ndarray): The pressure tensor in GPa generated per step by Lammps
        energy_kin (float): The kinetic energy of the system in eV (Default is None)
        time_step (float): MD time step in fs. (Default is 1 fs.)
        pressure_damping_timescale (float): Damping timescale in fs. (Default is None, no barostat is used.)
        compressibility (float): The compressibility of water in bar-1. More information here:
            http://www.sklogwiki.org/SklogWiki/index.php/Berendsen_barostat (Default is 4.57e-5 bar-1)
        structure (Atoms): The structure whose cell and positions are to be scaled.
        positions (numpy.ndarray): The updated positions from `VerletPositionUpdate`.
        previous_volume (float): The volume of the cell from the previous step in Ang3 (Default is None)
        pressure_style (string): 'isotorpic' or 'anisotropic'. (Default is 'anisotropic')
    Output attributes:
        pressure (float): The isotropic pressure in GPa
        volume (float): The volume of the cell in Ang3
        structure (Atoms): The scaled structure, corresponding to the simulated volume
        positions (numpy.ndarray): The scaled positions
    """

    def __init__(self, name=None):
        super(BerendsenBarostat, self).__init__(name=name)
        id_ = self.input.default
        id_.pressure = 0.
        id_.temperature = 0.
        id_.pressure_damping_timescale = 1000.
        id_.time_step = 1.
        id_.compressibility = 4.57e-5  # compressibility of water in bar^-1
        id_.pressure_style = 'isotropic'

    def command(self, pressure, temperature, box_pressure, energy_kin, time_step, positions,
                pressure_damping_timescale, compressibility, structure, previous_volume, pressure_style):

        if pressure_style != 'isotropic' and pressure_style != 'anisotropic':
            raise TypeError('style can only be \'isotropic\' or \'anisotropic\'')

        n_atoms = len(structure.positions)

        if previous_volume is None:
            previous_volume = structure.cell.diagonal().prod()

        if energy_kin is None:
            energy_kin = (3 / 2) * n_atoms * KB * temperature

        isotropic_pressure = np.trace(box_pressure) / 3  # pyiron stores pressure in GPa

        if pressure is None:
            new_structure = structure.copy()
            total_pressure = isotropic_pressure
        elif pressure is not None and pressure_style == 'isotropic':
            new_structure = structure.copy()
            new_structure.positions = positions
            first_term = ((2 * energy_kin) / (3 * previous_volume)) * EV_PER_ANGCUB_TO_GPA
            tau = (time_step / pressure_damping_timescale) * (compressibility / 3)
            total_pressure = first_term + isotropic_pressure  # GPa
            eta = 1 - (tau * (pressure - total_pressure) * GPA_TO_BAR)
            new_cell = new_structure.cell * eta
            new_structure.set_cell(new_cell, scale_atoms=True)
        elif pressure is not None and pressure_style == 'anisotropic':
            new_structure = structure.copy()
            new_structure.positions = positions
            first_term = ((2 * energy_kin) / (3 * previous_volume)) * EV_PER_ANGCUB_TO_GPA
            tau = (time_step / pressure_damping_timescale) * (compressibility / 3)
            total_pressure_x = first_term + box_pressure[0, 0]  # GPa
            eta_x = 1 - (tau * (pressure - total_pressure_x) * GPA_TO_BAR)
            total_pressure_y = first_term + box_pressure[1, 1]  # GPa
            eta_y = 1 - (tau * (pressure - total_pressure_y) * GPA_TO_BAR)
            total_pressure_z = first_term + box_pressure[2, 2]  # GPa
            eta_z = 1 - (tau * (pressure - total_pressure_z) * GPA_TO_BAR)

            old_cell = new_structure.cell
            new_cell = np.array([eta_x * old_cell[0],
                                 eta_y * old_cell[1],
                                 eta_z * old_cell[2]])

            new_structure.set_cell(new_cell, scale_atoms=True)
            total_pressure = np.mean([total_pressure_x, total_pressure_y, total_pressure_z])
        else:
            raise TypeError('Invalid value for pressure')

        return {
            'pressure': total_pressure,
            'structure': new_structure,
            'positions': new_structure.positions
        }


class ComputeFormationEnergy(PrimitiveVertex):
    """
    """

    def command(self, n_atoms, eq_energy, harm_to_inter_mean, harm_to_inter_std, harm_to_inter_se, inter_to_vac_mean,
                inter_to_vac_std, inter_to_vac_se):

        harm_to_inter_fe = unumpy.uarray(harm_to_inter_mean, harm_to_inter_std)
        harm_to_inter_fe_se = unumpy.uarray(harm_to_inter_mean, harm_to_inter_se)
        inter_to_vac_fe = unumpy.uarray(inter_to_vac_mean, inter_to_vac_std)
        inter_to_vac_fe_se = unumpy.uarray(inter_to_vac_mean, inter_to_vac_se)

        fe = ((eq_energy + harm_to_inter_fe) / n_atoms) + inter_to_vac_fe
        fe_se = ((eq_energy + harm_to_inter_fe_se) / n_atoms) + inter_to_vac_fe_se

        return {
            'formation_energy_mean': float(unumpy.nominal_values(fe)),
            'formation_energy_std': float(unumpy.std_devs(fe)),
            'formation_energy_se': float(unumpy.std_devs(fe_se))
        }



import xml.etree.ElementTree as ET

import numpy as np

from pyiron_base import DataContainer


class ARFitProperty(DataContainer):
    """
    Class to describe properties that can be fitted using atomicrex.
    Property and target value have to be given,
    the other parameters control details of the fitting procedure.
    For more information what they do visit the atomicrex documentation.
    
    Direct interaction with this class shouldn't be necessary.
    To conveniently create ARFitProperty objects use the
    ARFitPropertList add_fit_property method.
    """    
    def __init__(
            self,
            prop=None,
            target_value=None,
            fit=None,
            relax=None,
            relative_weight=None,
            residual_style=None,
            output=None,
            tolerance=None,
            min_val=None,
            max_val=None,
            output_all=None,
            *args,
            **kwargs
    ):
        super().__init__(table_name="fit_property", *args, **kwargs)
        self._prop = None
        self._residual_style = None
        if prop is not None:
            self.prop = prop
        if residual_style is not None:
            self.residual_style = residual_style
        self.target_value = target_value
        self.fit = fit
        self.relax = relax
        self.relative_weight = relative_weight
        self.tolerance = tolerance
        self.min_val = min_val
        self.max_val = max_val
        self.final_value = None
        self.output_all = output_all

    @property
    def prop(self):
        return self._prop

    @prop.setter
    def prop(self, prop):
        """Only allow properties that can be fitted using atomicrex

        Args:
            prop (string): name of the property

        Raises:
            ValueError: Given property can not be fitted using atomicrex.
        """    
        fittable_properties = [
            "atomic-energy",
            "atomic-forces",
            "bulk-modulus",
            "pressure",
        ]
        #cij_list = [f"c{i}{j}" for i, j in range(1,7) if j>=i]
        #fittable_properties.extend(cij_list)
        if prop in fittable_properties:
            self._prop = prop
        else:
            raise ValueError(f"prop should be one of {fittable_properties}")

    @property
    def residual_style(self):
        return self._residual_style

    @residual_style.setter
    def residual_style(self, residual_style):
        """Residual styles available in atomicrex

        Args:
            residual_style (string): [description]

        Raises:
            ValueError: [description]
        """        
        res_styles = ["squared", "squared-relative", "absolute-diff"]
        if residual_style in res_styles:
            self._residual_style = residual_style
        else:
            raise ValueError(f"residual style has to be one of {res_styles}")

    def _is_scalar(self):
        """
        Internal helper function, scalar properties are provided in the structure xml
        and vector properties (atomic forces) in the structure file
        Returns:
            [bool]: True if scalar, False if vector property.
        """        
        if self.prop != "atomic-forces":
            return True
        else:
            return False

    def to_xml_element(self):
        xml = ET.Element(f"{self.prop}")

        if self.fit:
            xml.set("fit", "true")
        else:
            xml.set("fit", "false")
        #xml.set("relax", f"{self.relax}".lower())
        xml.set("relative-weight", f"{self.relative_weight}")
        if self.tolerance is not None:
            xml.set("tolerance", f"{self.tolerance}")
        
        if self._is_scalar():
            xml.set("target", f"{self.target_value}")
            if self.min_val is not None:
                xml.set("min", f"{self.min_val}")
            if self.max_val is not None:
                xml.set("min", f"{self.max_val}")
            xml.set("residual-style", f"{self.residual_style}")
        else:
            if self.residual_style == "squared-relative":
                raise ValueError("Squared-relative residual style is not implemented for forces in atomicrex")
            if self.min_val is not None or self.max_val is not None:
                raise ValueError("Min and Max val can only be given for scalar properties")
            if self.output_all:
                xml.set("output-all", "true")
            xml.set("residual-style", f"{self.residual_style}")
        return xml

    @staticmethod
    def _parse_final_value(line):
        """
        Parses the final values of properties used in the fitting process from
        atomicrex output.

        Args:
            line (string): string from atomicrex output containing final value of some property

        Returns:
            [(string, float)]: property, final value
        """        
        if line.startswith("atomic-forces avg/max"):
            return "atomic-forces", None
        else:
            line = line.split()
            return line[0].rstrip(":"), float(line[1])


class ARFitPropertyList(DataContainer):
    """
    DataContainer of ARFitProperties that additionally provides utility functions
    that allow convenient addition of fit properties to a structure.
    Also provides internal functionality.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(table_name="fit_property", *args, **kwargs)


    def add_FitProperty(
            self,
            prop,
            target_value,
            fit=True,
            relax=False,
            relative_weight=1,
            residual_style="squared",
            output=True,
            tolerance=None,
            min_val=None,
            max_val=None,
            output_all=True,
    ):
        """
        Adds a fittable property to the fit properties
        of an atomicrex structure.
        Default values should be ok for most cases,
        but it is strongly recommended to check the purpose
        of each argument in the atomicrex documentation.

        Args:
            prop (string): property to be fitted
            target_value (float):
            fit (bool, optional): Include the property in the fit procedure. Defaults to True.
            relax (bool, optional): Calculate before or after structure relaxation. Defaults to False.
            relative_weight (int, optional): Weight for the objective function. Defaults to 1.
            residual_style (str, optional): See atomicrex documentation. Defaults to "squared-relative".
            output (bool, optional): Determines if the value is written to output.
            Could cause parsing problems if False. Defaults to True.
            tolerance (float, optional): See atomicrex documentation. Defaults to None.
            min_val (float, optional): Only scalar properties, if relaxation enabled. Defaults to None.
            max_val (float, optional): Only scalar properties, if relaxation enabled. Defaults to None.
            output_all (bool, optional): Only vector properties. Determines if full vector is written to output. Defaults to True.

        """    
        self[prop] = ARFitProperty(
            prop = prop,
            target_value = target_value,
            fit = fit,
            relax = relax,
            relative_weight = relative_weight,
            residual_style = residual_style,
            output = output,
            tolerance = tolerance,
            min_val = min_val,
            max_val = max_val,
            output_all=output_all,
        )

    def to_xml_element(self):
        """Internal helper function converting the list into an atomicrex xml element.
        """        
        properties = ET.Element("properties")
        for p in self.values():
            properties.append(p.to_xml_element)


Residual_Styles = [
"squared",
"squared-relative",
"absolute-diff"
]


class FlattenedARProperty:
    """
    Class to read and write scalar properties of a structure, f.e. the energy.
    """    
    def __init__(self, num_structures, prop):
        self.num_structures = num_structures
        self.prop = prop
        self._init_arrays()
    

    def _init_arrays(self):
        self.target_value = np.full(self.num_structures, np.nan)
        self.fit = np.full(self.num_structures, fill_value=False, dtype=bool)
        self.relative_weight = np.ones(self.num_structures) # default 1
        # resiudal styles are "squared", "squared-relative" and "absolute-diff"
        # Translate to 0, 1 and 2
        self.residual_style = np.zeros(self.num_structures, dtype=np.uint8) # default 0
        self.relax = np.full(self.num_structures, fill_value=False, dtype=bool) # default False
        self.tolerance = np.full(self.num_structures, fill_value=np.nan)
        self.min_val = np.full(self.num_structures, fill_value=np.nan)
        self.max_val = np.full(self.num_structures, fill_value=np.nan)
        self.final_value = np.full(self.num_structures, fill_value=np.nan)
        self.output = np.full(self.num_structures, fill_value=False, dtype=bool)

    def to_hdf(self, hdf, group_name=None):
        if group_name is None:
            group_name = f"{self.property}"
        with hdf.open(group_name) as h:
            h["target_value"] = self.target_value
            h["fit"] = self.fit
            h["relative_weight"] = self.relative_weight
            h["residual_style"] = self.residual_style
            h["relax"] = self.relax
            h["tolerance"] = self.tolerance
            h["min_val"] = self.min_val
            h["max_val"] = self.max_val
            h["final_value"] = self.final_value
            h["output"] = self.output

    #def _type_to_hdf(self, hdf):
    #    """
    #    Internal helper function to save type and version in hdf root#
    #
    #    Args:
    #        hdf (ProjectHDFio): HDF5 group object
    #    """
    #    hdf["TYPE"] = str(type(self))

    def from_hdf(self, hdf, group_name=None):
        if group_name is None:
            group_name = f"{self.prop}"
        with hdf.open(group_name) as h:
            self.target_value = h["target_value"]
            self.fit = h["fit"]
            self.relative_weight = h["relative_weight"]
            self.residual_style = h["residual_style"]
            self.relax = h["relax"]
            self.tolerance = h["tolerance"]
            self.min_val = h["min_val"]
            self.max_val = h["max_val"]
            self.output = h["output"]
            self.final_value = h["final_value"]

    def to_xml_element(self, index):
        xml = ET.Element(f"{self.prop}")
        if self.output[index]:
            xml.set("output", "true")
        if self.fit[index]:
            xml.set("fit", "true")
            xml.set("target", f"{self.target_value[index]}")
            #xml.set("relax", f"{self.relax}".lower())
            xml.set("relative-weight", f"{self.relative_weight[index]}")
            xml.set("residual-style", f"{Residual_Styles[self.residual_style[index]]}")
            if not np.isnan(self.tolerance[index]):
                xml.set("tolerance", f"{self.tolerance[index]}")
            if not np.isnan(self.min_val[index]):
                xml.set("min", f"{self.min_val[index]}")
            if not np.isnan(self.max_val[index]):
                xml.set("min", f"{self.max_val[index]}")
        return xml


class FlattenedARVectorProperty:
    """
    Like AR property, but for vector properties, i.e. forces
    """
    def __init__(self, num_structures, num_atoms, prop):
        self.num_structures = num_structures
        self.num_atoms = num_atoms
        self.prop = prop
        self._init_arrays()

    
    def _init_arrays(self):
        self.target_value = np.full((self.num_atoms, 3), np.nan)
        self.fit = np.full(self.num_structures, fill_value=False, dtype=bool)
        self.relative_weight = np.ones(self.num_structures) # default 1
        # resiudal styles are "squared", "squared-relative" and "absolute-diff"
        # Translate to 0, 1 and 2
        self.residual_style = np.zeros(self.num_structures, dtype=np.uint8) # default 0
        self.relax = np.full(self.num_structures, fill_value=False, dtype=bool) # default False
        self.tolerance = np.full(self.num_structures, fill_value=np.nan)
        self.final_value = np.full((self.num_atoms, 3), fill_value=np.nan)
        self.output = np.full(self.num_structures, fill_value=False, dtype=bool)

    def to_hdf(self, hdf, group_name=None):
        if group_name is None:
            group_name = f"{self.prop}"
        with hdf.open(group_name) as h:
            h["target_value"] = self.target_value
            h["fit"] = self.fit
            h["relative_weight"] = self.relative_weight
            h["residual_style"] = self.residual_style
            h["relax"] = self.relax
            h["tolerance"] = self.tolerance
            h["final_value"] = self.final_value
            h["output"] = self.output

    def from_hdf(self, hdf, group_name=None):
        if group_name is None:
            group_name = f"{self.prop}"
        with hdf.open(group_name) as h:
            self.target_value = h["target_value"]
            self.fit = h["fit"]
            self.relative_weight = h["relative_weight"]
            self.residual_style = h["residual_style"]
            self.relax = h["relax"]
            self.tolerance = h["tolerance"]
            self.final_value = h["final_value"]
            self.output = h["output"]
    
#    def _type_to_hdf(self, hdf):
#        """
#        Internal helper function to save type and version in hdf root
#
#        Args:
#            hdf (ProjectHDFio): HDF5 group object
#        """
#        hdf["TYPE"] = str(type(self))
    
    def to_xml_element(self, index):
        xml = ET.Element(f"{self.prop}")
        if self.output[index]:
            xml.set("output-all", "true")
        if self.fit[index]:
            xml.set("output-all", "true")
            xml.set("fit", "true")
            #xml.set("relax", f"{self.relax}".lower())
            xml.set("relative-weight", f"{self.relative_weight[index]}")
            xml.set("residual-style", f"{Residual_Styles[self.residual_style[index]]}")
            if not np.isnan(self.tolerance[index]):
                xml.set("tolerance", f"{self.tolerance[index]}")
        return xml




# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Demonstrator job for the RuNNer Neural Network Potential.
"""

import os.path
from glob import glob

from pyiron_base import GenericJob, Executable, Settings, DataContainer

import numpy as np
import pandas as pd

__author__ = "Marvin Poul"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "0.1"
__maintainer__ = "Marvin Poul"
__email__ = "poul@mpie.de"
__status__ = "development"
__date__ = "March 3, 2021"

s = Settings()

AngstromToBohr = 1.88972612
ElectronVoltToHartree = 0.03674932218


class RunnerFit(GenericJob):
    def __init__(self, project, job_name):
        super().__init__(project, job_name) 
        self._training_ids = []

        self.input = DataContainer(table_name="input")
        self.input.element = "Cu"

        s.publication_add(self.publication)

    @property
    def publication(self):
        return {
            "runner": [
                {
                    "title": "First Principles Neural Network Potentials for Reactive Simulations of Large Molecular and Condensed Systems",
                    "journal": "Angewandte Chemie International Edition",
                    "volume": "56",
                    "number": "42",
                    "year": "2017",
                    "issn": "1521-3773",
                    "doi": "10.1002/anie.201703114",
                    "url": "https://doi.org/10.1002/anie.201703114",
                    "author": ["Jörg Behler"],
                },
                {
                    "title": "Constructing high‐dimensional neural network potentials: A tutorial review",
                    "journal": "International Journal of Quantum Chemistry",
                    "volume": "115",
                    "number": "16",
                    "year": "2015",
                    "issn": "1097-461X",
                    "doi": "10.1002/qua.24890",
                    "url": "https://doi.org/10.1002/qua.24890",
                    "author": ["Jörg Behler"],
                },
                {
                    "title": "Generalized Neural-Network Representation of High-Dimensional Potential-Energy Surfaces",
                    "journal": "Physical Review Letters",
                    "volume": "98",
                    "number": "14",
                    "year": "2007",
                    "issn": "1079-7114",
                    "doi": "10.1103/PhysRevLett.98.146401",
                    "url": "https://doi.org/10.1103/PhysRevLett.98.146401",
                    "author": ["Jörg Behler", "Michelle Parrinello"],
                },
            ]
        }

    def add_job_to_fitting(self, job):
        """
        Add a job to the training database.  Currently only :class:`.TrainingContainer` are supported.

        Args:
            job (:class:`.TrainingContainer`): job to add to database
        """
        self._training_ids.append(job.id)

    def write_input(self):
        with open(os.path.join(self.working_directory, "input.data"), "w") as f:
            for id in self._training_ids:
                container = self.project.load(id)
                for atoms, energy, forces, _ in zip(*container.to_list()):
                    f.write("begin\n")
                    c = np.array(atoms.cell) * AngstromToBohr
                    f.write(f"lattice {c[0][0]:13.08f} {c[0][1]:13.08f} {c[0][2]:13.08f}\n")
                    f.write(f"lattice {c[1][0]:13.08f} {c[1][1]:13.08f} {c[1][2]:13.08f}\n")
                    f.write(f"lattice {c[2][0]:13.08f} {c[2][1]:13.08f} {c[2][2]:13.08f}\n")
                    p = atoms.positions * AngstromToBohr
                    ff = forces * ElectronVoltToHartree / AngstromToBohr
                    for i in range(len(atoms)):
                        f.write(f"atom {p[i, 0]:13.08f} {p[i, 1]:13.08f} {p[i, 2]:13.08f}")
                        f.write(f" {atoms.elements[i].Abbreviation} 0.0 0.0")
                        f.write(f" {ff[i, 0]:13.08f} {ff[i, 1]:13.08f} {ff[i, 2]:13.08f}\n")
                    f.write(f"energy {energy * ElectronVoltToHartree}\n")
                    f.write("charge 0.0\nend\n")

        with open(os.path.join(self.working_directory, "input.nn"), "w") as f:
            f.write(input_template.format(element=self.input.element))

    @property
    def lammps_potential(self):
        """
        :class:`pandas.DataFrame`: dataframe compatible with :attribute:`.LammpsInteractive.potential`
        """
        if not self.status.finished:
            raise RuntimeError("Job must have successfully finished before potential files can be copied!")
        weight_file = glob(f'{self.working_directory}/weights.*.data')[0]
        return pd.DataFrame({
                    'Name': ['RuNNer-Cu'],
                    'Filename': [[f'{self.working_directory}/input.nn',
                                  weight_file,
                                  f'{self.working_directory}/scaling.data']],
                    'Model': ['RuNNer'],
                    'Species': [['Cu']],
                    'Config': [['pair_style nnp dir "./" showew no showewsum 0 resetew no maxew 100 cflength 1.8897261328 cfenergy 0.0367493254 emap "1:Cu"\n',
                                'pair_coeff * * 12\n']]
                  })

    def collect_output(self):
        pass

    # To link to the executable from the notebook
    def _executable_activate(self, enforce=False):     
        if self._executable is None or enforce:
            self._executable = Executable(
                codename="runner", module="runner", path_binary_codes=s.resource_paths
            )

input_template = """### ####################################################################################################################
### This is the input file for the RuNNer tutorial (POTENTIALS WORKSHOP 2021-03-10) 
### This input file is intended for release version 1.2
### RuNNer is hosted at www.gitlab.com. The most recent version can only be found in this repository.
### For access please contact Prof. Jörg Behler, joerg.behler@uni-goettingen.de
###
### ####################################################################################################################
### General remarks: 
### - commands can be switched off by using the # character at the BEGINNING of the line
### - the input file can be structured by blank lines and comment lines
### - the order of the keywords is arbitrary
### - if keywords are missing, default values will be used and written to runner.out
### - if mandatory keywords or keyword options are missing, RuNNer will stop with an error message 
###
########################################################################################################################
########################################################################################################################
### The following keywords just represent a subset of the keywords offered by RuNNer
########################################################################################################################
########################################################################################################################

########################################################################################################################
### general keywords
########################################################################################################################
nn_type_short 1                           # 1=Behler-Parrinello
runner_mode 1                             # 1=calculate symmetry functions, 2=fitting mode, 3=predicition mode
number_of_elements 1                      # number of elements
elements {element}                               # specification of elements
random_seed 10                            # integer seed for random number generator                         
random_number_type 6                      # 6 recommended       

########################################################################################################################
### NN structure of the short-range NN  
########################################################################################################################
use_short_nn                              # use NN for short range interactions    
global_hidden_layers_short 2              # number of hidden layers               
global_nodes_short 15 15                  # number of nodes in hidden layers     
global_activation_short t t l             # activation functions  (t = hyperbolic tangent, l = linear)              

########################################################################################################################
### symmetry function generation ( mode 1): 
########################################################################################################################
test_fraction 0.10000                     # threshold for splitting between fitting and test set 

########################################################################################################################
### symmetry function definitions (all modes): 
########################################################################################################################
cutoff_type 1
symfunction_short {element}  2 {element}     0.000000      0.000000     12.000000
symfunction_short {element}  2 {element}     0.006000      0.000000     12.000000
symfunction_short {element}  2 {element}     0.016000      0.000000     12.000000
symfunction_short {element}  2 {element}     0.040000      0.000000     12.000000
symfunction_short {element}  2 {element}     0.109000      0.000000     12.000000

symfunction_short {element}  3 {element} {element}     0.00000       1.000000      1.000000     12.000000
symfunction_short {element}  3 {element} {element}     0.00000       1.000000      2.000000     12.000000
symfunction_short {element}  3 {element} {element}     0.00000       1.000000      4.000000     12.000000
symfunction_short {element}  3 {element} {element}     0.00000       1.000000     16.000000     12.000000
symfunction_short {element}  3 {element} {element}     0.00000      -1.000000      1.000000     12.000000
symfunction_short {element}  3 {element} {element}     0.00000      -1.000000      2.000000     12.000000
symfunction_short {element}  3 {element} {element}     0.00000      -1.000000      4.000000     12.000000
symfunction_short {element}  3 {element} {element}     0.00000      -1.000000     16.000000     12.000000

########################################################################################################################
### fitting (mode 2):general inputs for short range AND electrostatic part:
########################################################################################################################
epochs 20                                 # number of epochs                                     
fitting_unit eV                           # unit for error output in mode 2 (eV or Ha)
precondition_weights                      # optional precondition initial weights 

########################################################################################################################
### fitting options ( mode 2): short range part only:
########################################################################################################################
short_energy_error_threshold 0.10000      # threshold of adaptive Kalman filter short E         
short_force_error_threshold 1.00000       # threshold of adaptive Kalman filter short F        
kalman_lambda_short 0.98000               # Kalman parameter short E/F, do not change                        
kalman_nue_short 0.99870                  # Kalman parameter short E/F, do not change                      
use_short_forces                          # use forces for fitting                         
repeated_energy_update                    # optional: repeat energy update for each force update   
mix_all_points                            # do not change                    
scale_symmetry_functions                  # optional
center_symmetry_functions                 # optional 
short_force_fraction 0.01                 #
force_update_scaling -1.0                 #  

########################################################################################################################
### output options for mode 2 (fitting):  
########################################################################################################################
write_trainpoints                         # write trainpoints.out and testpoints.out files      
write_trainforces                         # write trainforces.out and testforces.out files    

########################################################################################################################
### output options for mode 3 (prediction):  
########################################################################################################################
calculate_forces                          # calculate forces    
calculate_stress                          # calculate stress tensor 
"""



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
from pyiron_contrib.protocol.generic import Vertex, PrimitiveVertex, CompoundVertex
from pyiron_contrib.protocol.utils import InputDictionary, Pointer
import numpy as np
import time
from abc import abstractmethod
from multiprocessing import Process, Manager
from pyiron_atomistics.vasp.interactive import VaspInteractive
from pyiron_atomistics.sphinx.interactive import SphinxInteractive

"""
A command class for running multiple of the same node
"""

__author__ = "Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Jul 23, 2019"


class ListVertex(PrimitiveVertex):
    """
    A base class for making wrappers to run multiple instances of the same protocol. Protocols which inherit from this
    class require the sub-protocol being copied to be passed in at initialization as an argument.

    Attributes to be assigned to the

    Attributes:
        child_type (Command): A class inheriting from ``Command`` with which to create child instances. (Passed as an
            argument at instantiation.)
        children (list): The instances of the child command to execute. These are created automatically (at run-time if
            they don't exist already).
        broadcast (InputDictionary): Input data to be split element-wise across all the child commands. (Entries here
            should always have the same number of entries as there are children, i.e. each child can have its own
            value.)
        direct (InputDictionary): Input data which is to be copied to each child (i.e. all children have the same
            value.)

    Input attributes:
        n_children (int): How many children to create.
    """

    def __init__(self, child_type):
        if not issubclass(child_type, Vertex):
            raise TypeError('ListVertex children must inherit from Protocol.')
        self.children = None  # Ahead of super so the n_history call doesn't trigger the setter and find no children
        super(ListVertex, self).__init__()
        self.child_type = child_type
        self._initialized = False
        self.direct = InputDictionary()
        self.broadcast = InputDictionary()
        self._n_history = None
        self.n_history = 1

    @abstractmethod
    def command(self, n_children):
        pass

    def finish(self):
        for child in self.children:
            child.finish()
        super(ListVertex, self).finish()

    def _initialize(self, n_children):
        children = [self.child_type(name="child_{}".format(n)) for n in range(n_children)]

        # Locate children in graph
        for n, child in enumerate(children):
            child.graph_parent = self
            child.vertex_name = "child_{}".format(n)

        # Link input to input.direct
        for key in list(self.direct.keys()):
            for child in children:
                setattr(child.input, key, getattr(Pointer(self.direct), key))

        # Link input.default to input.direct.default
        for key in list(self.direct.default.keys()):
            for child in children:
                setattr(child.input.default, key, getattr(Pointer(self.direct.default), key))

        # Link input to input.broadcast
        for key in list(self.broadcast.keys()):
            for n, child in enumerate(children):
                setattr(child.input, key, getattr(Pointer(self.broadcast), key)[n])

        # Link input.default to input.broadcast.default
        for key in list(self.broadcast.default.keys()):
            for n, child in enumerate(children):
                setattr(child.input.default, key, getattr(Pointer(self.broadcast.default), key)[n])

        self.children = children
        self._initialized = True

    @property
    def n_history(self):
        return self._n_history

    @n_history.setter
    def n_history(self, n_hist):
        self._n_history = n_hist
        if self.children is not None:
            for child in self.children:
                child.n_history = n_hist

    def _extract_output_data_from_children(self):
        output_keys = list(self.children[0].output.keys())  # Assumes that all the children are the same...
        if len(output_keys) > 0:
            output_data = {}
            for key in output_keys:
                values = []
                for child in self.children:
                    values.append(child.output[key][-1])
                output_data[key] = values
        else:
            output_data = None
        return output_data

    def to_hdf(self, hdf=None, group_name=None):
        super(ListVertex, self).to_hdf(hdf=hdf, group_name=group_name)
        hdf[group_name]['initialized'] = self._initialized
        if self.children is not None:
            with hdf.open(group_name + "/children") as hdf5_server:
                for n, child in enumerate(self.children):
                    child.to_hdf(hdf=hdf5_server, group_name="child" + str(n))

    def from_hdf(self, hdf=None, group_name=None):
        super(ListVertex, self).from_hdf(hdf=hdf, group_name=group_name)
        self._initialized = hdf[group_name]['initialized']
        if self._initialized:
            with hdf.open(group_name + "/children") as hdf5_server:
                children = []
                for n in np.arange(self.input.n_children, dtype=int):
                    child = self.child_type(name="child_{}".format(n))
                    child.from_hdf(hdf=hdf5_server, group_name="child" + str(n))
                    child.graph_parent = self
                    children.append(child)
            self.children = children


class ParallelList(ListVertex):
    """
    A list of commands which are executed in in parallel. The current implementation uses multiprocessing.Pool.

    Attributes:
        pool (multiprocessing.Pool): Define the number of workers that will be utilized to run the child jobs.
            Instantiation and closing are handled by default in `__init__` and `finish`.

        Note: It is best to set the number of workers to the number of cores so as to prevent larger computation
            times due to subprocess communication between the large number of workers in a single core.
    """

    def __init__(self, child_type, sleep_time=0):
        super(ParallelList, self).__init__(child_type)
        self.sleep_time = sleep_time

    def command(self, n_children):
        """This controls how the commands are run and is about logistics."""
        if self.children is None:
            self._initialize(n_children)

        for child in self.children:
            child.parallel_setup()

        start_time = time.time()
        sleep_time = ~self.sleep_time

        all_child_output = Manager().dict()

        jobs = []
        for i, child in enumerate(self.children):
            job = Process(target=child.execute_parallel, args=(i, all_child_output))
            job.start()
            time.sleep(sleep_time)
            jobs.append(job)

        for job in jobs:
            job.join()
            time.sleep(sleep_time)

        print(all_child_output.keys())

        ordered_child_output = dict.fromkeys(range(len(all_child_output)))
        for i in range(len(all_child_output)):
            ordered_child_output[i] = all_child_output[i]

        output_keys = list(ordered_child_output[0].keys())  # Assumes that all the children are the same...
        if len(output_keys) > 0:
            output_data = {}
            for key in output_keys:
                values = []
                for child in ordered_child_output.values():
                    values.append(child[key])
                output_data[key] = values
        else:
            output_data = None

        stop_time = time.time()
        print('Time elapsed :', stop_time - start_time)

        return output_data

    def finish(self):
        super(ParallelList, self).finish()


class SerialList(ListVertex):
    """
    A list of commands which are run in serial.
    """
    def __init__(self, child_type):
        super(SerialList, self).__init__(child_type=child_type)

    def command(self, n_children):
        """This controls how the commands are run and is about logistics."""
        if self.children is None:
            self._initialize(n_children)

        for child in self.children:
            child.execute()

        output_data = self._extract_output_data_from_children()
        return output_data


class AutoList(ParallelList, SerialList):
    """
    Choose between `SerialList` and `ParallelList` depending on the nature of the children.
    """
    def __init__(self, child_type):
        super(AutoList, self).__init__(child_type=child_type)

    def _is_expensive(self):
        try:
            if isinstance(
                self.children[0],
                (VaspInteractive, SphinxInteractive, CompoundVertex)
            ):  # Whitelist types that should be treated in parallel
                return True
            else:  # Everything else is cheap enough to treat in serial
                return False
        except (AttributeError, KeyError, TypeError):
            return False

    def command(self, n_children):
        if self._is_expensive():
            return ParallelList.command(self, n_children)
        else:
            return SerialList.command(self, n_children)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.


import sys

"""
this module provides a function to nicely print the whitelists
"""


__author__ = "Dominik Gehringer, Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Dec 12, 2019"


def count_paths(node):
    """
    Counts the entries in a nested dectionaries
    Args:
        node: (dict) the nested dictionary

    Returns: (int) the number of total leaf nodes

    """
    if not isinstance(node, dict):
        return 1
    else:
        s = 0
        for _, v in node.items():
            s += count_paths(v)
        return s


def print_tree(current_node, indent='', last='downup', name='root', file=sys.stdout):
    """
    Prints a nested dictionary in pretty tree
    This code is adapted version found in "https://github.com/clemtoy/pptree/blob/master/pptree/pptree.py"

    Args:
        current_node: (dict) as nested dictonaries, its keys will be converted to string in order to print the tree
        indent: (str) optional indentation
        last: (str) "downdown", "upup", "downup" or "updown" the direction of the last shape (default: "downup")
        name: (str) the name of the root node
        file: (file) location where to print the output (default: sys.stdout)

    Returns:

    """
    children = lambda node: [(k,v)
                             if isinstance(v, dict)
                             else ('%s: %s' %(k,v), None)
                             for k,v in node.items()]  if isinstance(node, dict) else []
    nb_children = lambda node: sum(count_paths(child) for childname, child in children(node))+1
    size_branch = {childname: nb_children(child) for (childname, child) in children(current_node)}
    # Creation of balanced lists for "up" branch and "down" branch
    up = sorted(children(current_node), key=lambda node: nb_children(node[0]))
    down = []
    while up and sum(size_branch[downname] for downname, downchild in down) < sum(size_branch[upname] for upname, upchild in up):
        down.append(up.pop())

    # print the "up" branch
    for childname, child in up:
        next_last = 'up' if up.index((childname, child)) == 0 else ''
        next_indent = '{0}{1}{2}'.format(indent, ' ' if 'up' in last else '│', ' ' * len(name))
        print_tree(child, next_indent, next_last, name=childname, file=file)

    # Printing of current node
    if last == 'up': start_shape = '┌'
    elif last == 'down': start_shape = '└'
    elif last == 'updown': start_shape = ' '
    else: start_shape = '├'

    if up: end_shape = '┤'
    elif down: end_shape = '┐'
    else: end_shape = ''

    print('{0}{1}{2}{3}'.format(indent, start_shape, name, end_shape), file=file)

    # Printing of "down" branch
    for childname, child in down:
        next_last = 'down' if down.index((childname, child)) is len(down) - 1 else ''
        next_indent = '{0}{1}{2}'.format(indent, ' ' if 'down' in last else '│', ' ' * len(name))
        print_tree(child, next_indent, next_last, name=childname, file=file)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Alternative structure container that stores them in flattened arrays.
"""

from itertools import chain

import numpy as np
import h5py

from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_atomistics.atomistics.structure.has_structure import HasStructure

class StructureStorage(HasStructure):
    """
    Class that can write and read lots of structures from and to hdf quickly.

    This is done by storing positions, cells, etc. into large arrays instead of writing every structure into a new
    group.  Structures are stored together with an identifier that should be unique.  The class can be initialized with
    the number of structures and the total number of atoms in all structures, but re-allocates memory as necessary when
    more (or larger) structures are added than initially anticipated.

    You can add structures and a human-readable name with :method:`.add_structure()`.

    >>> container = StructureStorage()
    >>> container.add_structure(Atoms(...), "fcc")
    >>> container.add_structure(Atoms(...), "hcp")
    >>> container.add_structure(Atoms(...), "bcc")

    Accessing stored structures works with :method:`.get_strucure()`.  You can either pass the identifier you passed
    when adding the structure or the numeric index

    >>> container.get_structure(frame=0) == container.get_structure(frame="fcc")
    True

    Custom arrays may also be defined on the container

    >>> container.add_array("energy", shape=(), dtype=np.float64, fill=-1, per="structure")

    You can then pass arrays of the corresponding shape to :method:`add_structure()`

    >>> container.add_structure(Atoms(...), "grain_boundary", energy=3.14)

    Saved arrays are accessed with :method:`.get_array()`

    >>> container.get_array("energy", 3)
    3.14
    >>> container.get_array("energy", 0)
    -1

    It is also possible to use the same names in :method:`.get_array()` as in :method:`.get_structure()`.

    >>> container.get_array("energy", 0) == container.get_array("energy", "fcc")
    True

    The length of the container is the number of structures inside it.

    >>> len(container)
    4

    By default the following arrays are defined for each structure:
        - identifier    shape=(),    dtype=str,          per structure; human readable name of the structure
        - cell          shape=(3,3), dtype=np.float64,   per structure; cell shape
        - pbc           shape=(3,),  dtype=bool          per structure; periodic boundary conditions
        - symbols:      shape=(),    dtype=str,          per atom; chemical symbol
        - positions:    shape=(3,),  dtype=np.float64,   per atom: atomic positions
    If a structure has spins/magnetic moments defined on its atoms these will be saved in a per atom array as well.  In
    that case, however all structures in the container must either have all collinear spins or all non-collinear spins.

    When adding new array follow the convention that per-structure arrays should be named in singular and per-atom
    arrays should be named in plural.
    """

    __version__ = "0.1.0"
    __hdf_version__ = "0.1.0"

    def __init__(self, num_structures=1, num_atoms=1):
        """
        Create new structure container.

        Args:
            num_structures (int): pre-allocation for per structure arrays
            num_atoms (int): pre-allocation for per atoms arrays
        """
        # tracks allocated versed as yet used number of structures/atoms
        self._num_structures_alloc = self.num_structures = num_structures
        self._num_atoms_alloc = self.num_atoms = num_atoms
        # store the starting index for properties with unknown length
        self.current_atom_index = 0
        # store the index for properties of known size, stored at the same index as the structure
        self.current_structure_index = 0
        # Also store indices of structure recently added
        self.prev_structure_index = 0
        self.prev_atom_index = 0

        self._init_arrays()

    def __len__(self):
        return self.current_structure_index

    def _init_arrays(self):
        self._per_atom_arrays = {
                # 2 character unicode array for chemical symbols
                "symbols": np.full(self._num_atoms_alloc, "XX", dtype=np.dtype("U2")),
                "positions": np.empty((self._num_atoms_alloc, 3))
        }

        self._per_structure_arrays = {
                "start_index": np.empty(self._num_structures_alloc, dtype=np.int32),
                "length": np.empty(self._num_structures_alloc, dtype=np.int32),
                "identifier": np.empty(self._num_structures_alloc, dtype=np.dtype("U20")),
                "cell": np.empty((self._num_structures_alloc, 3, 3)),
                "pbc": np.empty((self._num_atoms_alloc, 3), dtype=bool)
        }

    @property
    def symbols(self):
        """:meta private:"""
        return self._per_atom_arrays["symbols"]

    @property
    def positions(self):
        """:meta private:"""
        return self._per_atom_arrays["positions"]

    @property
    def start_index(self):
        """:meta private:"""
        return self._per_structure_arrays["start_index"]

    @property
    def length(self):
        """:meta private:"""
        return self._per_structure_arrays["length"]

    @property
    def identifier(self):
        """:meta private:"""
        return self._per_structure_arrays["identifier"]

    @property
    def cell(self):
        """:meta private:"""
        return self._per_structure_arrays["cell"]

    @property
    def pbc(self):
        """:meta private:"""
        return self._per_structure_arrays["pbc"]

    def get_elements(self):
        """
        Return a list of chemical elements in the training set.

        Returns:
            :class:`list`: list of unique elements in the training set as strings of their standard abbreviations
        """
        return list(set(self._per_atom_arrays["symbols"]))

    def _get_per_atom_slice(self, frame):
        start = self._per_structure_arrays["start_index"][frame]
        end = start + self._per_structure_arrays["length"][frame]
        return slice(start, end, 1)

    def _resize_atoms(self, new):
        self._num_atoms_alloc = new
        for k, a in self._per_atom_arrays.items():
            new_shape = (new,) + a.shape[1:]
            try:
                a.resize(new_shape)
            except ValueError:
                self._per_atom_arrays[k] = np.resize(a, new_shape)

    def _resize_structures(self, new):
        self._num_structures_alloc = new
        for k, a in self._per_structure_arrays.items():
            new_shape = (new,) + a.shape[1:]
            try:
                a.resize(new_shape)
            except ValueError:
                self._per_structure_arrays[k] = np.resize(a, new_shape)

    def add_array(self, name, shape=(), dtype=np.float64, fill=None, per="atom"):
        """
        Add a custom array to the container.

        When adding an array after some structures have been added, specifying `fill` will be used as a default value
        for the value of the array for those structures.

        Adding an array with the same name twice is ignored, if dtype and shape match, otherwise raises an exception.

        >>> container = StructureStorage()
        >>> container.add_structure(Atoms(...), "foo")
        >>> container.add_array("energy", shape=(), dtype=np.float64, fill=42, per="structure")
        >>> container.get_array("energy", 0)
        42

        Args:
            name (str): name of the new array
            shape (tuple of int): shape of the new array per atom or structure; scalars can pass ()
            dtype (type): data type of the new array, string arrays can pass 'U$n' where $n is the length of the string
            fill (object): populate the new array with this value for existing structure, if given; default `None`
            per (str): either "atom" or "structure"; denotes whether the new array should exist for every atom in a
                       structure or only once for every structure; case-insensitive

        Raises:
            ValueError: if wrong value for `per` is given
            ValueError: if array with same name but different parameters exists already
        """

        if name in self._per_atom_arrays:
            a = self._per_atom_arrays[name]
            if a.shape[1:] != shape or a.dtype != dtype or per != "atom":
                raise ValueError(f"Array with name '{name}' exists with shape {a.shape[1:]} and dtype {a.dtype}.")
            else:
                return

        if name in self._per_structure_arrays:
            a = self._per_structure_arrays[name]
            if a.shape[1:] != shape or a.dtype != dtype or per != "structure":
                raise ValueError(f"Array with name '{name}' exists with shape {a.shape[1:]} and dtype {a.dtype}.")
            else:
                return

        per = per.lower()
        if per == "atom":
            shape = (self._num_atoms_alloc,) + shape
            store = self._per_atom_arrays
        elif per == "structure":
            shape = (self._num_structures_alloc,) + shape
            store = self._per_structure_arrays
        else:
            raise ValueError(f"per must \"atom\" or \"structure\", not {per}")

        if fill is None:
            store[name] = np.empty(shape=shape, dtype=dtype)
        else:
            store[name] = np.full(shape=shape, fill_value=fill, dtype=dtype)

    def get_array(self, name, frame):
        """
        Fetch array for given structure.

        Works for per atom and per arrays.

        Args:
            name (str): name of the array to fetch
            frame (int, str): selects structure to fetch, as in :method:`.get_structure()`

        Returns:
            :class:`numpy.ndarray`: requested array

        Raises:
            `KeyError`: if array with name does not exists
        """

        if isinstance(frame, str):
            frame = self._translate_frame(frame)
        if name in self._per_atom_arrays:
            return self._per_atom_arrays[name][self._get_per_atom_slice(frame)]
        elif name in self._per_structure_arrays:
            return self._per_structure_arrays[name][frame]
        else:
            raise KeyError(f"no array named {name} defined on StructureStorage")

    def set_array(self, name, frame, value):
        """
        Add array for given structure.

        Works for per atom and per arrays.

        Args:
            name (str): name of array to set
            frame (int, str): selects structure to set, as in :method:`.get_strucure()`

        Raises:
            `KeyError`: if array with name does not exists
        """

        if isinstance(frame, str):
            frame = self._translate_frame(frame)
        if name in self._per_atom_arrays:
            self._per_atom_arrays[name][self._get_per_atom_slice(frame)] = value
        elif name in self._per_structure_arrays:
            self._per_structure_arrays[name][frame] = value
        else:
            raise KeyError(f"no array named {name} defined on StructureStorage")

    def has_array(self, name):
        """
        Checks whether an array of the given name exists and returns meta data given to :method:`.add_array()`.


        >>> container.has_array("energy")
        {'shape': (), 'dtype': np.float64, 'per': 'atom'}
        >>> container.has_array("fnorble")
        None

        Args:
            name (str): name of the array to check

        Returns:
            None: if array does not exist
            dict: if array exists, 
        """
        if name in self._per_atom_arrays:
            a = self._per_atom_arrays[name]
            per = "atom"
        elif name in self._per_structure_arrays:
            a = self._per_structure_arrays[name]
            per = "structure"
        else:
            return None
        return {"shape": a.shape[1:], "dtype": a.dtype, "per": per}


    def add_structure(self, structure, identifier=None, **arrays):
        """
        Add a new structure to the container.

        Additional keyword arguments given specify additional arrays to store for the structure.  If an array with the
        given keyword name does not exist yet, it will be added to the container.

        >>> container = StructureStorage()
        >>> container.add_structure(Atoms(...), identifier="A", energy=3.14)
        >>> container.get_array("energy", 0)
        3.14

        If the first axis of the extra array matches the length of the given structure, it will be added as an per atom
        array, otherwise as an per structure array.

        >>> structure = Atoms(...)
        >>> container.add_structure(structure, identifier="B", forces=len(structure) * [[0,0,0]])
        >>> len(container.get_array("forces", 1)) == len(structure)
        True

        Reshaping the array to have the first axis be length 1 forces the array to be set as per structure array.  That
        axis will then be stripped.

        >>> container.add_structure(Atoms(...), identifier="C", pressure=np.eye(3)[np.newaxis, :, :])
        >>> container.get_array("pressure", 2).shape
        (3, 3)

        Args:
            structure (:class:`.Atoms`): structure to add
            identifier (str, optional): human-readable name for the structure, if None use current structre index as
                                        string
            **kwargs: additional arrays to store for structure
        """

        if identifier is None:
            identifier = str(self.num_structures)

        n = len(structure)
        new_atoms = self.current_atom_index + n

        if new_atoms > self._num_atoms_alloc:
            self._resize_atoms(max(new_atoms, self._num_atoms_alloc * 2))
        if self.current_structure_index + 1 > self._num_structures_alloc:
            self._resize_structures(self._num_structures_alloc * 2)

        if new_atoms > self.num_atoms:
            self.num_atoms = new_atoms
        if self.current_structure_index + 1 > self.num_structures:
            self.num_structures += 1

        # len of structure to index into the initialized arrays
        i = self.current_atom_index + n

        self._per_atom_arrays["symbols"][self.current_atom_index:i] = np.array(structure.symbols)
        self._per_atom_arrays["positions"][self.current_atom_index:i] = structure.positions

        self._per_structure_arrays["start_index"][self.current_structure_index] = self.current_atom_index
        self._per_structure_arrays["length"][self.current_structure_index] = n
        self._per_structure_arrays["identifier"][self.current_structure_index] = identifier
        self._per_structure_arrays["cell"][self.current_structure_index] = structure.cell.array
        self._per_structure_arrays["pbc"][self.current_structure_index] = structure.pbc

        if structure.spins is not None:
            arrays["spins"] = structure.spins

        for k, a in arrays.items():
            a = np.asarray(a)
            if len(a.shape) > 0 and a.shape[0] == n:
                if k not in self._per_atom_arrays:
                    self.add_array(k, shape=a.shape[1:], dtype=a.dtype, per="atom")
                self._per_atom_arrays[k][self.current_atom_index:i] = a
            else:
                if len(a.shape) > 0 and a.shape[0] == 1:
                    a = a[0]
                if k not in self._per_structure_arrays:
                    self.add_array(k, shape=a.shape, dtype=a.dtype, per="structure")
                self._per_structure_arrays[k][self.current_structure_index] = a

        self.prev_structure_index = self.current_structure_index
        self.prev_atom_index = self.current_atom_index

        # Set new current_atom_index and increase current_structure_index
        self.current_structure_index += 1
        self.current_atom_index = i
        #return last_structure_index, last_atom_index


    def _type_to_hdf(self, hdf):
        """
        Internal helper function to save type and version in hdf root

        Args:
            hdf (ProjectHDFio): HDF5 group object
        """
        hdf["NAME"] = self.__class__.__name__
        hdf["TYPE"] = str(type(self))
        hdf["VERSION"] = self.__version__
        hdf["HDF_VERSION"] = self.__hdf_version__
        hdf["OBJECT"] = "StructureStorage"

    def to_hdf(self, hdf, group_name="structures"):
        # truncate arrays to necessary size before writing
        self._resize_atoms(self.num_atoms)
        self._resize_structures(self.num_structures)

        with hdf.open(group_name) as hdf_s_lst:
            self._type_to_hdf(hdf_s_lst)
            hdf_s_lst["num_atoms"] =  self._num_atoms_alloc
            hdf_s_lst["num_structures"] = self._num_structures_alloc

            hdf_arrays = hdf_s_lst.open("arrays")
            for k, a in chain(self._per_atom_arrays.items(), self._per_structure_arrays.items()):
                if a.dtype.char == "U":
                    # numpy stores unicode data in UTF-32/UCS-4, but h5py wants UTF-8, so we manually encode them here
                    # TODO: string arrays with shape != () not handled
                    hdf_arrays[k] = np.array([s.encode("utf8") for s in a],
                                             # each character in a utf8 string might be encoded in up to 4 bytes, so to
                                             # make sure we can store any string of length n we tell h5py that the
                                             # string will be 4 * n bytes; numpy's dtype does this calculation already
                                             # in itemsize, so we don't need to repeat it here
                                             # see also https://docs.h5py.org/en/stable/strings.html
                                             dtype=h5py.string_dtype('utf8', a.dtype.itemsize))
                else:
                    hdf_arrays[k] = a


    def from_hdf(self, hdf, group_name="structures"):
        with hdf.open(group_name) as hdf_s_lst:
            version = hdf_s_lst.get("HDF_VERSION", "0.0.0")
            self._num_structures_alloc = self.num_structures = self.current_structure_index = hdf_s_lst["num_structures"]
            self._num_atoms_alloc = self.num_atoms = self.current_atom_index = hdf_s_lst["num_atoms"]

            if version == "0.1.0":
                with hdf_s_lst.open("arrays") as hdf_arrays:
                    for k in hdf_arrays.list_nodes():
                        a = np.array(hdf_arrays[k])
                        if a.dtype.char == "S":
                            # if saved as bytes, we wrote this as an encoded unicode string, so manually decode here
                            # TODO: string arrays with shape != () not handled
                            a = np.array([s.decode("utf8") for s in a],
                                        # itemsize of original a is four bytes per character, so divide by four to get
                                        # length of the orignal stored unicode string; np.dtype('U1').itemsize is just a
                                        # platform agnostic way of knowing how wide a unicode charater is for numpy
                                        dtype=f"U{a.dtype.itemsize//np.dtype('U1').itemsize}")
                        if a.shape[0] == self._num_atoms_alloc:
                            self._per_atom_arrays[k] = a
                        elif a.shape[0] == self._num_structures_alloc:
                            self._per_structure_arrays[k] = a

            elif version == "0.0.0":
                self._per_atom_arrays["symbols"] = hdf_s_lst["symbols"].astype(np.dtype("U2"))
                self._per_atom_arrays["positions"] = hdf_s_lst["positions"]

                self._per_structure_arrays["start_index"] = hdf_s_lst["start_indices"]
                self._per_structure_arrays["length"] = hdf_s_lst["len_current_struct"]
                self._per_structure_arrays["identifier"] = hdf_s_lst["identifiers"].astype(np.dtype("U20"))
                self._per_structure_arrays["cell"] = hdf_s_lst["cells"]

                self._per_structure_arrays["pbc"] = np.full((self.num_structures, 3), True)


    def _translate_frame(self, frame):
        for i, name in enumerate(self._per_structure_arrays["identifier"]):
            if name == frame:
                return i
        raise KeyError(f"No structure named {frame} in StructureStorage.")

    def _get_structure(self, frame=-1, wrap_atoms=True):
        slc = self._get_per_atom_slice(frame)
        if "spins" in self._per_atom_arrays:
            magmoms = self._per_atom_arrays["spins"][slc]
        else:
            # not all structures have spins saved on them
            magmoms = None
        return Atoms(symbols=self._per_atom_arrays["symbols"][slc],
                     positions=self._per_atom_arrays["positions"][slc],
                     cell=self._per_structure_arrays["cell"][frame],
                     pbc=self._per_structure_arrays["pbc"][frame],
                     magmoms=magmoms)

    def _number_of_structures(self):
        return len(self)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Store structures together with energies and forces for potential fitting applications.

Basic usage:

>>> pr = Project("training")
>>> container = pr.create.job.TrainingContainer("small_structures")

Let's make a structure and invent some forces

>>> structure = pr.create.structure.ase_bulk("Fe")
>>> forces = numpy.array([-1, 1, -1])
>>> container.include_structure(structure, energy=-1.234, forces=forces, name="Fe_bcc")

If you have a lot of precomputed structures you may also add them in bulk from a pandas DataFrame

>>> df = pandas.DataFrame({ "name": "Fe_bcc", "atoms": structure, "energy": -1.234, "forces": forces })
>>> container.include_dataset(df)

You can retrieve the full database with :method:`~.TrainingContainer.to_pandas()` like this

>>> container.to_pandas()
name    atoms   energy  forces  number_of_atoms
Fe_bcc  ...
"""

from warnings import catch_warnings

import numpy as np
import pandas as pd
from pyiron_contrib.atomistics.atomistics.job.structurestorage import StructureStorage
from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_atomistics.atomistics.structure.has_structure import HasStructure
from pyiron_base import GenericJob


class TrainingContainer(GenericJob, HasStructure):
    """
    Stores ASE structures with energies and forces.
    """

    def __init__(self, project, job_name):
        super().__init__(project=project, job_name=job_name)
        self.__name__ = "TrainingContainer"
        self.__hdf_version__ = "0.2.0"
        self._container = StructureStorage()
        self._container.add_array("energy", dtype=np.float64, per="structure")
        self._container.add_array("forces", shape=(3,), dtype=np.float64, per="atom")
        self._table_cache = None

    @property
    def _table(self):
        if self._table_cache is None:
            self._table_cache = pd.DataFrame({
                "name":             [self._container.get_array("identifier", i)
                                        for i in range(len(self._container))],
                "atoms":            [self._container.get_structure(i)
                                        for i in range(len(self._container))],
                "energy":           [self._container.get_array("energy", i)
                                        for i in range(len(self._container))],
                "forces":           [self._container.get_array("forces", i)
                                        for i in range(len(self._container))],
            })
            self._table_cache["number_of_atoms"] = [len(s) for s in self._table_cache.atoms]
        return self._table_cache

    def include_job(self, job, iteration_step=-1):
        """
        Add structure, energy and forces from job.

        Args:
            job (:class:`.AtomisticGenericJob`): job to take structure from
            iteration_step (int, optional): if job has multiple steps, this selects which to add
        """
        self.include_structure(job.get_structure(iteration_step=iteration_step),
                               energy=job.output.energy_pot[iteration_step],
                               forces=job.output.forces[iteration_step],
                               name=job.name)

    def include_structure(self, structure, energy, forces=None, name=None):
        """
        Add new structure to structure list and save energy and forces with it.

        For consistency with the rest of pyiron, energy should be in units of eV and forces in eV/A, but no conversion
        is performed.

        Args:
            structure_or_job (:class:`~.Atoms`): structure to add
            energy (float): energy of the whole structure
            forces (Nx3 array of float, optional): per atom forces, where N is the number of atoms in the structure
            name (str, optional): name describing the structure
        """
        if forces is not None:
            self._container.add_structure(structure, name, energy=energy, forces=forces)
        else:
            self._container.add_structure(structure, name, energy=energy)
        if self._table_cache:
            self._table = self._table.append(
                    {"name": name, "atoms": structure, "energy": energy, "forces": forces,
                     "number_of_atoms": len(structure)},
                    ignore_index=True)

    def include_dataset(self, dataset):
        """
        Add a pandas DataFrame to the saved structures.

        The dataframe should have the following columns:
            - name: human readable name of the structure
            - atoms(:class:`ase.Atoms`): the atomic structure
            - energy(float): energy of the whole structure
            - forces (Nx3 array of float): per atom forces, where N is the number of atoms in the structure
        """
        self._table_cache = self._table.append(dataset, ignore_index=True)
        # in case given dataset has more columns than the necessary ones, swallow/ignore them in *_
        for name, atoms, energy, forces, *_ in dataset.itertuples(index=False):
            self._container.add_structure(atoms, name, energy=energy, forces=forces)

    def _get_structure(self, frame=-1, wrap_atoms=True):
        return self._container.get_structure(frame=frame, wrap_atoms=wrap_atoms)

    def _number_of_structures(self):
        return self._container.number_of_structures

    def get_elements(self):
        """
        Return a list of chemical elements in the training set.

        Returns:
            :class:`list`: list of unique elements in the training set as strings of their standard abbreviations
        """
        return self._container.get_elements()

    def to_pandas(self):
        """
        Export list of structure to pandas table for external fitting codes.

        The table contains the following columns:
            - 'name': human-readable name of the structure
            - 'ase_atoms': the structure as a :class:`.Atoms` object
            - 'energy': the energy of the full structure
            - 'forces': the per atom forces as a :class:`numpy.ndarray`, shape Nx3
            - 'number_of_atoms': the number of atoms in the structure, N

        Returns:
            :class:`pandas.DataFrame`: collected structures
        """
        return self._table

    def to_list(self, filter_function=None):
        """
        Returns the data as lists of pyiron structures, energies, forces, and the number of atoms

        Args:
            filter_function (function): Function applied to the dataset (which is a pandas DataFrame) to filter it

        Returns:
            tuple: list of structures, energies, forces, and the number of atoms
        """
        if filter_function is None:
            data_table = self._table
        else:
            data_table = filter_function(self._table)
        structure_list = data_table.atoms.to_list()
        energy_list = data_table.energy.to_list()
        force_list = data_table.forces.to_list()
        num_atoms_list = data_table.number_of_atoms.to_list()
        return structure_list, energy_list, force_list, num_atoms_list

    def write_input(self):
        pass

    def collect_output(self):
        pass

    def run_static(self):
        self.status.finished = True

    def run_if_interactive(self):
        self.to_hdf()
        self.status.finished = True

    def to_hdf(self, hdf=None, group_name=None):
        super().to_hdf(hdf=hdf, group_name=group_name)
        self._container.to_hdf(self.project_hdf5, "structures")

    def from_hdf(self, hdf=None, group_name=None):
        super().from_hdf(hdf=hdf, group_name=group_name)
        hdf_version = self.project_hdf5.get("HDF_VERSION", "0.1.0")
        if hdf_version == "0.1.0":
            table = pd.read_hdf(self.project_hdf5.file_name, self.name + "/output/structure_table")
            self.include_dataset(table)
        else:
            self._container = StructureStorage()
            self._container.from_hdf(self.project_hdf5, "structures")



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import os
from pyiron.lammps.base import Input
from pyiron.lammps.interactive import LammpsInteractive
from pyiron_contrib.atomistics.mlip.mlip import read_cgfs
from pyiron_base import GenericParameters

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class LammpsMlip(LammpsInteractive):
    def __init__(self, project, job_name):
        super(LammpsMlip, self).__init__(project, job_name)
        self.input = MlipInput()
        self.__name__ = "LammpsMlip"
        self.__version__ = None  # Reset the version number to the executable is set automatically
        self._executable = None
        self._executable_activate()

    def set_input_to_read_only(self):
        """
        This function enforces read-only mode for the input classes, but it has to be implement in the individual
        classes.
        """
        super(LammpsMlip, self).set_input_to_read_only()
        self.input.mlip.read_only = True

    def write_input(self):
        super(LammpsMlip, self).write_input()
        if self.input.mlip['mlip:load-from'] == 'auto':
            self.input.mlip['mlip:load-from'] = os.path.basename(self.potential['Filename'][0][0])
        self.input.mlip.write_file(file_name="mlip.ini", cwd=self.working_directory)

    def enable_active_learning(self):
        self.input.mlip.load_string("""\
abinitio void
mlip mtpr
mlip:load-from Trained.mtp_
calculate-efs TRUE
fit FALSE
select TRUE
select:site-en-weight 0.0
select:energy-weight 1.0
select:force-weight 0.0
select:stress-weight 0.0
select:threshold-init 1e-5
select:threshold 2.0
select:threshold-swap 1.000001
select:threshold-break 5.0
select:save-selected selected.cfg
select:save-state selection.mvs
select:load-state state.mvs
select:efs-ignored FALSE
select:log selection.log
write-cfgs:skip 0
log lotf.log""")

    def collect_output(self):
        super(LammpsMlip, self).collect_output()
        if 'select:save-selected' in self.input.mlip._dataset['Parameter']:
            file_name = os.path.join(self.working_directory, self.input.mlip['select:save-selected'])
            if os.path.exists(file_name):
                cell, positions, forces, stress, energy, indicies, grades, jobids, timesteps = read_cgfs(file_name=file_name)
                with self.project_hdf5.open("output/mlip") as hdf5_output:
                    hdf5_output['forces'] = forces
                    hdf5_output['energy_tot'] = energy
                    hdf5_output['pressures'] = stress
                    hdf5_output['cells'] = cell
                    hdf5_output['positions'] = positions
                    hdf5_output['indicies'] = indicies


class MlipInput(Input):
    def __init__(self):
        self.mlip = MlipParameter()
        super(MlipInput, self).__init__()

    def to_hdf(self, hdf5):
        """
        
        Args:
            hdf5:
        Returns:
        """
        with hdf5.open("input") as hdf5_input:
            self.mlip.to_hdf(hdf5_input)
        super(MlipInput, self).to_hdf(hdf5)

    def from_hdf(self, hdf5):
        """
        
        Args:
            hdf5:
        Returns:
        """
        with hdf5.open("input") as hdf5_input:
            self.mlip.from_hdf(hdf5_input)
        super(MlipInput, self).from_hdf(hdf5)


class MlipParameter(GenericParameters):
    def __init__(self, separator_char=' ', comment_char='#', table_name="mlip_inp"):
        super(MlipParameter, self).__init__(separator_char=separator_char, comment_char=comment_char, table_name=table_name)
        
    def load_default(self, file_content=None):
        if file_content is None:
            file_content = '''\
abinitio void
mlip mtpr
mlip:load-from auto
calculate-efs TRUE
fit FALSE
select FALSE
'''
        self.load_string(file_content)




# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.


import threading
import logging
import inspect
from types import FunctionType, MethodType

"""
Provides classes for a custom more feature rich signal-slot system than PyQt4's. Allows slot deletion, and explicit
handler execution.
"""

__author__ = "Dominik Gehringer, Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "December 10, 2019"

class Synchronization(object):
    """
    Helper class which creates a mutex.
    """

    def __init__(self):
        super(Synchronization, self).__init__()
        self.mutex = threading.RLock()


class Event(Synchronization):
    """
    A which implements a signal slot system. It executes event handlers synchronized. An event handler may be
    :obj:`EventHandler` object, :obj:`function` or a :obj:`lambda`. Something :obj:`callable` at least.

    If you pass a :obj:`callable` as event handler it must have a ``__name__`` attribute. The handlers name will be
    ``eventHandler.name`` if it is a :obj:`EventHandler` object or a ``functions.__name__`` attribute.

    Lambdas do not get an identifier. So you cannot remove them except by :func:`clear_handlers`,
    but this resets all handlers.

    Attributes:
        suppressed (bool): A flag that indicates if the event object should block the fire calls. If set to
            `True` signals will be blocked.

    """

    def __init__(self):
        super(Event, self).__init__()
        self.__handlers = {}
        self.__unnamed_handlers = []
        self.supressed = False
        self.__synchronize(
            names=['add_event_handler', 'remove_event_handler', '__add__', '__sub__', 'handler_count', '__getitem__'])

    @property
    def handler_count(self):
        """
        Counts how many handlers are registered

        Returns:
            int: The total number of handlers
        """
        return len(self.__handlers) + len(self.__unnamed_handlers)

    @property
    def handler_keys(self):
        """
        `Property:` Creates a list with the names of all handlers

        Returns:
            :obj:`list` of :obj:`str`: a list with all handler names
        """
        return self.__handlers.keys()

    @property
    def named_handlers(self):
        """
        `Property:` All handlers which are registered with a name

        Returns:
             :obj:`list`: a list of all named handlers (handler with an identifier)
        """
        return self.__handlers.items()

    @property
    def unnamed_handlers(self):
        """
       `Property:` List of all lambdas

        Returns:
            :obj:`list` of :obj:`function`: a list of all registered lambdas
        """
        return self.__unnamed_handlers

    def add_event_handler(self, handler):
        """
        Adds a event handler to event. Each handler is identified by a string. The handlers identifier will be
        ``eventHandler.name`` if it is a :obj:`EventHandler` object or a ``functions.__name__`` attribute.
        Lambdas do not get an identifier. To fire this handler explicitly use::

            event.fire_handler(identifier)

        You can shortcut this function by using::

            event += lambda args: do_something(arg)
            event += EventHandler('my_handler_name', my_awesome_handler_func)
            event += my_object.my_awesome_handler_func

        Args:
            handler (:obj:`EventHandler`, :obj:`function` or :obj:`lambda`): the callable handler object

        Raises:
            TypeError: If handler is no instance of :obj:`EventHandler` a :obj:`function` or a :obj:`lambda`
            KeyError: If handlers identifier is already registered
        """
        if callable(handler):
            if hasattr(handler, '__name__'):
                name = handler.__name__
                if name == '<lambda>':
                    self.__unnamed_handlers.append(handler)
                    return
            else:
                self.__unnamed_handlers.append(handler)
                return
        elif isinstance(handler, EventHandler):
            name = handler.name
        else:
            raise TypeError(
                'The argument of add_event_handler must be of type event.EventHandler, function or a lambda expression')
        if name not in self.__handlers.keys():
            self.__handlers[name] = handler
        else:
            raise KeyError('The event already contains a handler named "{0}"'.format(name))

    def clear_handlers(self):
        """
        Removes all handlers, even registered :obj:`lambda` s
        """
        self.__handlers = {}
        self.__unnamed_handlers = []

    def fire(self, *args, **kwargs):
        """
        Causes all registered handlers to execute

        Note:
            The arguments specified in `*args` and `**kwargs` must be consistent with the handlers signatures

        Args:
            *args: Arguments passed to the handler functions
            **kwargs: Keyword arguments passed to the handler functions
        """
        self.mutex.acquire()
        try:
            local_copy = list(self.__handlers.values()) + self.__unnamed_handlers
        finally:
            self.mutex.release()
        if not self.supressed:
            for handler in local_copy:
                try:
                    if callable(handler):
                        if hasattr(handler, '__name__'):
                            name = handler.__name__
                        else:
                            name = str(handler)
                        handler(*args, **kwargs)
                    else:
                        name = handler.name
                        handler.func(*args, **kwargs)
                except:
                    logging.getLogger(__name__).exception(
                        'An error ocurred while executing function {func}.'.format(func=name), exc_info=True)

    def fire_handler(self, handler_name, *args, **kwargs):
        """
        Fires only a specific registered hander. Only handlers with a name can be fired explicitly.

        Theses handlers may be :obj:`EventHandler` objects or named :obj:`functions`

        Args:
            handler_name (:obj:`str`): The identifier of the handler
                If the underlying event handler is of type :obj:`EventHandler` the name is ``event_handler_obj.name``
                In case of a :obj:`function` the identifier is ``my_handler_func.__name__``
            *args: Arguments passed to the handler function
            **kwargs: Keyword arguments passed to the handler function
        """
        self.mutex.acquire()
        try:
            handler = self[handler_name]
        finally:
            self.mutex.release()
        if not self.supressed:
            try:
                if callable(handler):
                    if hasattr(handler, '__name__'):
                        name = handler.__name__
                    else:
                        name = str(handler)
                    handler(*args, **kwargs)
                else:
                    name = handler.name
                    handler.func(*args, **kwargs)
            except:
                logging.getLogger(__name__).exception(
                    'An error ocurred while executing function {func}.'.format(func=name), exc_info=True)

    def has_handler(self, handler):
        """
        Method to check if a certain event handler is already registered at the event

        Args:
            handler (:obj:`function`, :obj:`str` or :obj:`EventHandler`):  The handler or its name to check

        Returns:
            bool: `True` if handler is available else `False`
        """
        if callable(handler):
            if hasattr(handler, '__name__'):
                name = handler.__name__
            else:
                name = str(handler)
        elif isinstance(handler, EventHandler):
            name = handler.name
        elif isinstance(handler, str):
            name = handler
        return name in self.__handlers.keys()

    def remove_event_handler(self, handler):
        """
        Removes the given event handler from the event

        You can also shortcut this function by using::

            event -= 'my_handlers_identifier_string'
            event -= my_event_handler_object
            event -= my_object.my_awesome_handler_func

        Note:
            You cannot remove :obj:`lambda` expressions explicitly.

        Args:
            handler (:obj:`EventHandler`, :obj:`function` or :obj:`lambda`): the callable handler object

        Raises:
            TypeError: If handler is no instance of :obj:`EventHandler` a :obj:`function` or a :obj:`lambda`
            KeyError: If handlers name is not registered
        """
        if isinstance(handler, EventHandler):
            name = handler.name
        elif isinstance(handler, str):
            name = handler
        elif isinstance(handler, (FunctionType, MethodType)) or inspect.ismethod(handler):
            name = handler.__name__
        else:
            raise TypeError('The argument of remove_event_handler must be of type event. EventHandler, function or a '
                            'lambda expression')
        if name in self.__handlers.keys():
            del self.__handlers[name]
        else:
            raise KeyError('The event does not contain a handler named "{0}"'.format(name))

    def set_event_handler(self, handler):
        """
        Reassigns a event handler

        Args:
            handler (:obj:`EventHandler`, :obj:`function` or :obj:`lambda`): the callable handler object

        Raises:
            TypeError: If handler is no instance of :obj:`EventHandler` a :obj:`function` or a :obj:`lambda`
            KeyError: If handlers name is not registered
        """
        if callable(handler):
            if hasattr(handler, '__name__'):
                name = handler.__name__
                if name == '<lambda>':
                    raise KeyError('A lambda expression cannot be reassigned')
            else:
                raise KeyError('Only a named handler can be reassigned')
        elif isinstance(handler, EventHandler):
            name = handler.name
        else:
            raise TypeError(
                'The argument of set_event_handler must be of type event.EventHandler, function or a lambda expression')
        self.__handlers[name] = handler

    def __synchronize(self, names=None):
        """Synchronize methods in the given class.
        Only synchronize the methods whose names are
        given, or all methods if names=None."""
        for (name, val) in self.__dict__.items():
            if callable(val) and name != '__init__' and \
                    (names == None or name in names):
                # print("synchronizing", name)
                setattr(self, name, synchronized(val))

    def __add__(self, other):
        self.add_event_handler(other)
        return self

    def __getitem__(self, item):
        return self.__handlers[item]

    def __sub__(self, other):
        self.remove_event_handler(other)
        return self


class EventHandler:
    """
    Utility class to identify a callback classs with a name.

    Attributes:
        name (:obj:`str`): The name of the event handler
        func (:obj:`function`): The function which will be executed
    """

    def __init__(self, name, func):
        self.name = name
        self.func = func

    def __eq__(self, other):
        return self.name == other.name and type(self) == type(other)

    def __hash__(self):
        return self.name.__hash__()

    def __repr__(self):
        return 'event.EventHandler(name={0}, func={1})'.format(self.name, self.func)


def synchronized(name):
    """
    Function that creates a lock object and stores in the callers __dict__. Wrappes method for synchronized execution
    :param name: name of the callable to wrap

    Returns:
        :obj:`function`: the wrapped function
    """

    def synchronized_method(method):
        outer_lock = threading.Lock()
        lock_name = "__" + method.__name__ + "_lock_" + name + "__"

        def sync_method(self, *args, **kws):
            with outer_lock:
                if not hasattr(self, lock_name): setattr(self, lock_name, threading.Lock())
                lock = getattr(self, lock_name)
                with lock:
                    return method(self, *args, **kws)

        return sync_method

    return synchronized_method
